import pyglet
from . import util


class PhysicalObject(pyglet.sprite.Sprite):
    """A sprite with physical properties such as velocity"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Velocity
        self.velocity_x, self.velocity_y = 0.0, 0.0

        # Flags to toggle collision with bullets
        self.reacts_to_bullets = True
        self.is_bullet = False

        # Flag to remove this object from the game_object list
        self.dead = False

        # List of new objects to go in the game_objects list
        self.new_objects = []

        # Tell the game handler about any event handlers
        # Only applies to things with keyboard/mouse input
        self.event_handlers = []

    def update(self, dt):
        """This method should be called every frame."""

        # Update position according to velocity and time
        self.x += self.velocity_x * dt
        self.y += self.velocity_y * dt

        # Wrap around the screen if necessary
        self.check_bounds()

    def check_bounds(self):
        """Use the classic Asteroids screen wrapping behavior"""
        min_x = -self.image.width / 2
        min_y = -self.image.height / 2
        max_x = 800 + self.image.width / 2
        max_y = 600 + self.image.height / 2
        if self.x < min_x:
            self.x = max_x
        if self.y < min_y:
            self.y = max_y
        if self.x > max_x:
            self.x = min_x
        if self.y > max_y:
            self.y = min_y

    def collides_with(self, other_object):
        """Determine if this object collides with another"""

        # Ignore bullet collisions if we're supposed to
        if not self.reacts_to_bullets and other_object.is_bullet:
            return False
        if self.is_bullet and not other_object.reacts_to_bullets:
            return False

        # Calculate distance between object centers that would be a collision,
        # assuming square resources
        collision_distance = self.image.width * 0.5 * self.scale \
                             + other_object.image.width * 0.5 * other_object.scale

        # Get distance using position tuples
        actual_distance = util.distance(self.position, other_object.position)

        return (actual_distance <= collision_distance)

    def handle_collision_with(self, other_object):
        if other_object.__class__ is not self.__class__:
            self.dead = True



from __future__ import annotations

from ctypes import CFUNCTYPE, POINTER, c_ubyte, cast, create_string_buffer, pointer
from typing import Any, Callable, Sequence

import pyglet.lib
from pyglet.gl.lib import decorate_function, missing_function
from pyglet.util import asbytes

__all__ = ['link_GL', 'link_GLX']

gl_lib = pyglet.lib.load_library('GL')

# Look for glXGetProcAddressARB extension, use it as fallback (for ATI fglrx and DRI drivers).
try:
    glXGetProcAddressARB = getattr(gl_lib, 'glXGetProcAddressARB')  # noqa: B009, N816
    glXGetProcAddressARB.restype = POINTER(CFUNCTYPE(None))
    glXGetProcAddressARB.argtypes = [POINTER(c_ubyte)]
    _have_getprocaddress = True
except AttributeError:
    _have_getprocaddress = False


def link_GL(name: str, restype: Any, argtypes: Any, requires: str | None = None,  # noqa: N802
            suggestions: Sequence[str] | None = None) -> Callable[..., Any]:
    """Attempt to retrieve the GL function from the loaded library.

    If the function is not found, an attempt will be made via glXGetProcAddressARB.

    If both are unsuccessful, a dummy function will be returned that raises a MissingFunctionException.
    """
    try:
        func = getattr(gl_lib, name)
        func.restype = restype
        func.argtypes = argtypes
        decorate_function(func, name)
        return func
    except AttributeError:
        if _have_getprocaddress:
            # Fallback if implemented but not in ABI
            bname = cast(pointer(create_string_buffer(asbytes(name))), POINTER(c_ubyte))
            addr = glXGetProcAddressARB(bname)
            if addr:
                ftype = CFUNCTYPE(*((restype, *tuple(argtypes))))
                func = cast(addr, ftype)
                decorate_function(func, name)
                return func

    return missing_function(name, requires, suggestions)


link_GLX = link_GL  # noqa: N816



"""
Tests for the default application event loop.
"""
from threading import Event, Thread

from pyglet.app import event_loop
from tests import mock


def check_running():
    assert event_loop.is_running


def test_start_stop(performance):
    event_loop.clock.schedule_once(lambda dt: check_running(), .1)
    event_loop.clock.schedule_once(lambda dt: event_loop.exit(), .2)
    with performance.timer(1.):
        event_loop.run()
    assert not event_loop.is_running


def test_multiple_start_stop(performance):
    with performance.timer(30.):
        for _ in range(100):
            test_start_stop(performance)


def test_events():
    enter_mock = mock.MagicMock()
    exit_mock = mock.MagicMock()
    event_loop.push_handlers(on_enter=enter_mock,
                             on_exit=exit_mock)
    try:
        event_loop.clock.schedule_once(lambda dt: event_loop.exit(), .1)
        event_loop.run()
        enter_mock.assert_called_once_with()
        exit_mock.assert_called_once_with()
    finally:
        event_loop.pop_handlers()


def test_on_window_close():
    event_loop.clock.schedule_once(lambda dt: event_loop.on_window_close(None), .1)
    event_loop.run()
    assert not event_loop.is_running


def test_sleep(performance):
    def _sleep():
        event_loop.sleep(100.)
        _sleep.returned.set()
    _sleep.returned = Event()
    thread = Thread(target=_sleep)

    event_loop.clock.schedule_once(lambda dt: thread.start(), .1)
    event_loop.clock.schedule_once(lambda dt: event_loop.exit(), .2)
    with performance.timer(1.):
        event_loop.run()
    assert not event_loop.is_running
    assert _sleep.returned.wait(1.)



from ctypes import *

from pyglet.gl import *
from pyglet.image import *
from pyglet.image.codecs import *
from pyglet.image.codecs import gif

import pyglet.lib
import pyglet.window

gdk = pyglet.lib.load_library('gdk-x11-2.0')
gdkpixbuf = pyglet.lib.load_library('gdk_pixbuf-2.0')

GdkPixbufLoader = c_void_p
GdkPixbuf = c_void_p
guchar = c_char
gdkpixbuf.gdk_pixbuf_loader_new.restype = POINTER(GdkPixbufLoader)
gdkpixbuf.gdk_pixbuf_loader_get_pixbuf.restype = POINTER(GdkPixbuf)
gdkpixbuf.gdk_pixbuf_get_pixels.restype = POINTER(guchar)
gdkpixbuf.gdk_pixbuf_loader_get_animation.restype = POINTER(c_void_p)
gdkpixbuf.gdk_pixbuf_animation_get_iter.restype = POINTER(c_void_p)
gdkpixbuf.gdk_pixbuf_animation_iter_get_pixbuf.restype = POINTER(GdkPixbuf)


class GTimeVal(Structure):
    _fields_ = [
        ('tv_sec', c_long),
        ('tv_usec', c_long)
    ]


GQuark = c_uint32
gint = c_int
gchar = c_char


class GError(Structure):
    _fields_ = [
            ('domain', GQuark),
            ('code', gint),
            ('message', POINTER(gchar))
    ]

gerror_ptr = POINTER(GError)

def _gerror_to_string(error):
    """
    Convert a GError to a string.
    `error` should be a valid pointer to a GError struct.
    """
    return 'GdkPixBuf Error: domain[{}], code[{}]: {}'.format(error.contents.domain,
                                                              error.contents.code,
                                                              error.contents.message)


class GdkPixBufLoader:
    """
    Wrapper around GdkPixBufLoader object.
    """
    def __init__(self, filename, file):
        self.closed = False
        self._file = file
        self._filename = filename
        self._loader = gdkpixbuf.gdk_pixbuf_loader_new()
        if self._loader is None:
            raise ImageDecodeException('Unable to instantiate gdk pixbuf loader')
        self._load_file()

    def __del__(self):
        if self._loader is not None:
            if not self.closed:
                self._cancel_load()
            gdk.g_object_unref(self._loader)

    def _load_file(self):
        self._file.seek(0)
        data = self._file.read()
        self.write(data)

    def _finish_load(self):
        assert not self.closed
        error = gerror_ptr()
        all_data_passed = gdkpixbuf.gdk_pixbuf_loader_close(self._loader, byref(error))
        self.closed = True
        if not all_data_passed:
            raise ImageDecodeException(_gerror_to_string(error))

    def _cancel_load(self):
        assert not self.closed
        gdkpixbuf.gdk_pixbuf_loader_close(self._loader, None)
        self.closed = True

    def write(self, data):
        assert not self.closed, 'Cannot write after closing loader'
        error = gerror_ptr()
        if not gdkpixbuf.gdk_pixbuf_loader_write(self._loader, data, len(data), byref(error)):
            raise ImageDecodeException(_gerror_to_string(error))

    def get_pixbuf(self):
        self._finish_load()
        pixbuf = gdkpixbuf.gdk_pixbuf_loader_get_pixbuf(self._loader)
        if pixbuf is None:
            raise ImageDecodeException('Failed to get pixbuf from loader')
        return GdkPixBuf(self, pixbuf)

    def get_animation(self):
        self._finish_load()
        anim = gdkpixbuf.gdk_pixbuf_loader_get_animation(self._loader)
        if anim is None:
            raise ImageDecodeException('Failed to get animation from loader')
        gif_delays = self._get_gif_delays()
        return GdkPixBufAnimation(self, anim, gif_delays)

    def _get_gif_delays(self):
        # GDK pixbuf animations will loop indefinitely if looping is enabled for the
        # gif, so get number of frames and delays from gif metadata
        assert self._file is not None
        self._file.seek(0)
        gif_stream = gif.read(self._file)
        return [image.delay for image in gif_stream.images]


class GdkPixBuf:
    """
    Wrapper around GdkPixBuf object.
    """
    def __init__(self, loader, pixbuf):
        # Keep reference to loader alive
        self._loader = loader
        self._pixbuf = pixbuf
        gdk.g_object_ref(pixbuf)

    def __del__(self):
        if self._pixbuf is not None:
            gdk.g_object_unref(self._pixbuf)

    def load_next(self):
        return self._pixbuf is not None

    @property
    def width(self):
        assert self._pixbuf is not None
        return gdkpixbuf.gdk_pixbuf_get_width(self._pixbuf)

    @property
    def height(self):
        assert self._pixbuf is not None
        return gdkpixbuf.gdk_pixbuf_get_height(self._pixbuf)

    @property
    def channels(self):
        assert self._pixbuf is not None
        return gdkpixbuf.gdk_pixbuf_get_n_channels(self._pixbuf)

    @property
    def rowstride(self):
        assert self._pixbuf is not None
        return gdkpixbuf.gdk_pixbuf_get_rowstride(self._pixbuf)

    @property
    def has_alpha(self):
        assert self._pixbuf is not None
        return gdkpixbuf.gdk_pixbuf_get_has_alpha(self._pixbuf) == 1

    def get_pixels(self):
        pixels = gdkpixbuf.gdk_pixbuf_get_pixels(self._pixbuf)
        assert pixels is not None
        buf = (c_ubyte * (self.rowstride * self.height))()
        memmove(buf, pixels, self.rowstride * (self.height - 1) + self.width * self.channels)
        return buf

    def to_image(self):
        if self.width < 1 or self.height < 1 or self.channels < 1 or self.rowstride < 1:
            return None

        pixels = self.get_pixels()

        # Determine appropriate GL type
        if self.channels == 3:
            format = 'RGB'
        else:
            format = 'RGBA'

        return ImageData(self.width, self.height, format, pixels, -self.rowstride)


class GdkPixBufAnimation:
    """
    Wrapper for a GdkPixBufIter for an animation.
    """
    def __init__(self, loader, anim, gif_delays):
        self._loader = loader
        self._anim = anim
        self._gif_delays = gif_delays
        gdk.g_object_ref(anim)

    def __del__(self):
        if self._anim is not None:
            gdk.g_object_unref(self._anim)

    def __iter__(self):
        time = GTimeVal(0, 0)
        anim_iter = gdkpixbuf.gdk_pixbuf_animation_get_iter(self._anim, byref(time))
        return GdkPixBufAnimationIterator(self._loader, anim_iter, time, self._gif_delays)

    def to_animation(self):
        return Animation(list(self))


class GdkPixBufAnimationIterator:
    def __init__(self, loader, anim_iter, start_time, gif_delays):
        self._iter = anim_iter
        self._first = True
        self._time = start_time
        self._loader = loader
        self._gif_delays = gif_delays
        self.delay_time = None

    def __del__(self):
        if self._iter is not None:
            gdk.g_object_unref(self._iter)
        # The pixbuf returned by the iter is owned by the iter, so no need to destroy that one

    def __iter__(self):
        return self

    def __next__(self):
        self._advance()
        frame = self.get_frame()
        if frame is None:
            raise StopIteration
        return frame

    def _advance(self):
        if not self._gif_delays:
            raise StopIteration
        self.delay_time = self._gif_delays.pop(0)

        if self._first:
            self._first = False
        else:
            if self.gdk_delay_time == -1:
                raise StopIteration
            else:
                gdk_delay = self.gdk_delay_time * 1000 # milliseconds to microseconds
                us = self._time.tv_usec + gdk_delay
                self._time.tv_sec += us // 1000000
                self._time.tv_usec = us % 1000000
                gdkpixbuf.gdk_pixbuf_animation_iter_advance(self._iter, byref(self._time))

    def get_frame(self):
        pixbuf = gdkpixbuf.gdk_pixbuf_animation_iter_get_pixbuf(self._iter)
        if pixbuf is None:
            return None
        image = GdkPixBuf(self._loader, pixbuf).to_image()
        return AnimationFrame(image, self.delay_time)

    @property
    def gdk_delay_time(self):
        assert self._iter is not None
        return gdkpixbuf.gdk_pixbuf_animation_iter_get_delay_time(self._iter)


class GdkPixbuf2ImageDecoder(ImageDecoder):
    def get_file_extensions(self):
        return ['.png', '.xpm', '.jpg', '.jpeg', '.tif', '.tiff', '.pnm',
                '.ras', '.bmp', '.gif']

    def get_animation_file_extensions(self):
        return ['.gif', '.ani']

    def decode(self, filename, file):
        if not file:
            file = open(filename, 'rb')
        loader = GdkPixBufLoader(filename, file)
        return loader.get_pixbuf().to_image()

    def decode_animation(self, filename, file):
        if not file:
            file = open(filename, 'rb')
        loader = GdkPixBufLoader(filename, file)
        return loader.get_animation().to_animation()


def get_decoders():
    return [GdkPixbuf2ImageDecoder()]


def get_encoders():
    return []


def init():
    gdk.g_type_init()


init()



import pyglet

window = pyglet.window.Window()
label = pyglet.text.Label('Hello, world!',
                          font_size=36,
                          x=window.width // 2,
                          y=window.height // 2,
                          anchor_x='center',
                          anchor_y='center')


@window.event
def on_draw():
    window.clear()
    label.draw()


pyglet.app.run()



from __future__ import annotations

from ctypes import byref, c_int, c_uint, sizeof

from pyglet import gl
from pyglet.display.win32 import Win32Canvas
from pyglet.gl import gl_info, wgl, wgl_info, wglext_arb
from pyglet.libs.win32 import PIXELFORMATDESCRIPTOR, _gdi32
from pyglet.libs.win32.constants import (
    PFD_DEPTH_DONTCARE,
    PFD_DOUBLEBUFFER,
    PFD_DOUBLEBUFFER_DONTCARE,
    PFD_DRAW_TO_WINDOW,
    PFD_STEREO,
    PFD_STEREO_DONTCARE,
    PFD_SUPPORT_OPENGL,
    PFD_TYPE_RGBA,
)

from .base import DisplayConfig, Config, Context


class Win32Config(Config):  # noqa: D101

    def match(self, canvas: Win32Canvas) -> list[Win32DisplayConfig] | list[Win32DisplayConfigARB]:
        if not isinstance(canvas, Win32Canvas):
            msg = 'Canvas must be instance of Win32Canvas'
            raise RuntimeError(msg)

        # Use ARB API if available
        if gl_info.have_context() and wgl_info.have_extension('WGL_ARB_pixel_format'):
            return self._get_arb_pixel_format_matching_configs(canvas)

        return self._get_pixel_format_descriptor_matching_configs(canvas)

    def _get_pixel_format_descriptor_matching_configs(self, canvas: Win32Canvas) -> list[Win32DisplayConfig]:
        """Get matching configs using standard PIXELFORMATDESCRIPTOR technique."""
        pfd = PIXELFORMATDESCRIPTOR()
        pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR)
        pfd.nVersion = 1
        pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL

        if self.double_buffer:
            pfd.dwFlags |= PFD_DOUBLEBUFFER
        else:
            pfd.dwFlags |= PFD_DOUBLEBUFFER_DONTCARE

        if self.stereo:
            pfd.dwFlags |= PFD_STEREO
        else:
            pfd.dwFlags |= PFD_STEREO_DONTCARE

        #  Not supported in pyglet API: swap_copy: PFD_SWAP_COPY and swap_exchange: PFD_SWAP_EXCHANGE

        if not self.depth_size:
            pfd.dwFlags |= PFD_DEPTH_DONTCARE

        pfd.iPixelType = PFD_TYPE_RGBA
        pfd.cColorBits = self.buffer_size or 0
        pfd.cRedBits = self.red_size or 0
        pfd.cGreenBits = self.green_size or 0
        pfd.cBlueBits = self.blue_size or 0
        pfd.cAlphaBits = self.alpha_size or 0
        pfd.cAccumRedBits = self.accum_red_size or 0
        pfd.cAccumGreenBits = self.accum_green_size or 0
        pfd.cAccumBlueBits = self.accum_blue_size or 0
        pfd.cAccumAlphaBits = self.accum_alpha_size or 0
        pfd.cDepthBits = self.depth_size or 0
        pfd.cStencilBits = self.stencil_size or 0
        pfd.cAuxBuffers = self.aux_buffers or 0

        pf = _gdi32.ChoosePixelFormat(canvas.hdc, byref(pfd))
        if pf:
            return [Win32DisplayConfig(canvas, pf, self)]
        else:
            return []

    def _get_arb_pixel_format_matching_configs(self, canvas: Win32Canvas) -> list[Win32DisplayConfigARB]:
        """Get configs using the WGL_ARB_pixel_format extension.

        This method assumes a (dummy) GL context is already created.
        """
        # Check for required extensions
        if (self.sample_buffers or self.samples) and not gl_info.have_extension('GL_ARB_multisample'):
            return []

        # Construct array of attributes
        attrs = []
        for name, value in self.get_gl_attributes():
            attr = Win32DisplayConfigARB.attribute_ids.get(name, None)
            if attr and value is not None:
                attrs.extend([attr, int(value)])
        attrs.append(0)
        attrs = (c_int * len(attrs))(*attrs)

        pformats = (c_int * 16)()
        nformats = c_uint(16)
        wglext_arb.wglChoosePixelFormatARB(canvas.hdc, attrs, None, nformats, pformats, nformats)

        formats = [Win32DisplayConfigARB(canvas, pf, self) for pf in pformats[:nformats.value]]
        return formats


class Win32DisplayConfig(DisplayConfig):  # noqa: D101
    def __init__(self, canvas: Win32Canvas, pf: int, config: Win32Config) -> None:  # noqa: D107
        super().__init__(canvas, config)
        self._pf = pf
        self._pfd = PIXELFORMATDESCRIPTOR()

        _gdi32.DescribePixelFormat(canvas.hdc, pf, sizeof(PIXELFORMATDESCRIPTOR), byref(self._pfd))

        self.double_buffer = bool(self._pfd.dwFlags & PFD_DOUBLEBUFFER)
        self.sample_buffers = 0
        self.samples = 0
        self.stereo = bool(self._pfd.dwFlags & PFD_STEREO)
        self.buffer_size = self._pfd.cColorBits
        self.red_size = self._pfd.cRedBits
        self.green_size = self._pfd.cGreenBits
        self.blue_size = self._pfd.cBlueBits
        self.alpha_size = self._pfd.cAlphaBits
        self.accum_red_size = self._pfd.cAccumRedBits
        self.accum_green_size = self._pfd.cAccumGreenBits
        self.accum_blue_size = self._pfd.cAccumBlueBits
        self.accum_alpha_size = self._pfd.cAccumAlphaBits
        self.depth_size = self._pfd.cDepthBits
        self.stencil_size = self._pfd.cStencilBits
        self.aux_buffers = self._pfd.cAuxBuffers

    def compatible(self, canvas: Win32Canvas) -> bool:
        return isinstance(canvas, Win32Canvas)

    def create_context(self, share: Win32Context | None) -> Win32Context:
        return Win32Context(self, share)

    def _set_pixel_format(self, canvas: Win32Canvas) -> None:
        _gdi32.SetPixelFormat(canvas.hdc, self._pf, byref(self._pfd))


class Win32DisplayConfigARB(DisplayConfig):  # noqa: D101
    attribute_ids = {  # noqa: RUF012
        'double_buffer': wglext_arb.WGL_DOUBLE_BUFFER_ARB,
        'stereo': wglext_arb.WGL_STEREO_ARB,
        'buffer_size': wglext_arb.WGL_COLOR_BITS_ARB,
        'aux_buffers': wglext_arb.WGL_AUX_BUFFERS_ARB,
        'sample_buffers': wglext_arb.WGL_SAMPLE_BUFFERS_ARB,
        'samples': wglext_arb.WGL_SAMPLES_ARB,
        'red_size': wglext_arb.WGL_RED_BITS_ARB,
        'green_size': wglext_arb.WGL_GREEN_BITS_ARB,
        'blue_size': wglext_arb.WGL_BLUE_BITS_ARB,
        'alpha_size': wglext_arb.WGL_ALPHA_BITS_ARB,
        'depth_size': wglext_arb.WGL_DEPTH_BITS_ARB,
        'stencil_size': wglext_arb.WGL_STENCIL_BITS_ARB,
        'accum_red_size': wglext_arb.WGL_ACCUM_RED_BITS_ARB,
        'accum_green_size': wglext_arb.WGL_ACCUM_GREEN_BITS_ARB,
        'accum_blue_size': wglext_arb.WGL_ACCUM_BLUE_BITS_ARB,
        'accum_alpha_size': wglext_arb.WGL_ACCUM_ALPHA_BITS_ARB,
    }

    def __init__(self, canvas: Win32Canvas, pf: int, config: Win32Config) -> None:  # noqa: D107
        super().__init__(canvas, config)
        self._pf = pf

        names = list(self.attribute_ids.keys())
        attrs = list(self.attribute_ids.values())
        attrs = (c_int * len(attrs))(*attrs)
        values = (c_int * len(attrs))()

        wglext_arb.wglGetPixelFormatAttribivARB(canvas.hdc, pf, 0, len(attrs), attrs, values)

        for name, value in zip(names, values):
            setattr(self, name, value)

    def compatible(self, canvas: Win32Canvas) -> bool:
        # TODO more careful checking
        return isinstance(canvas, Win32Canvas)

    def create_context(self, share: Win32ARBContext | None) -> Win32ARBContext | Win32Context:
        if wgl_info.have_extension('WGL_ARB_create_context'):
            # Graphics adapters that ONLY support up to OpenGL 3.1/3.2
            # should be using the Win32ARBContext class.
            return Win32ARBContext(self, share)

        return Win32Context(self, share)

    def _set_pixel_format(self, canvas: Win32Canvas) -> None:
        _gdi32.SetPixelFormat(canvas.hdc, self._pf, None)


class _BaseWin32Context(Context):
    def __init__(self, config: Win32DisplayConfig | Win32DisplayConfigARB, share: Win32Context | Win32ARBContext) -> None:
        super().__init__(config, share)
        self._context = None

    def set_current(self) -> None:
        if self._context is not None and self != gl.current_context:
            wgl.wglMakeCurrent(self.canvas.hdc, self._context)
        super().set_current()

    def detach(self) -> None:
        if self.canvas:
            wgl.wglDeleteContext(self._context)
            self._context = None
        super().detach()

    def flip(self) -> None:
        _gdi32.SwapBuffers(self.canvas.hdc)

    def get_vsync(self) -> bool:
        if wgl_info.have_extension('WGL_EXT_swap_control'):
            return bool(wglext_arb.wglGetSwapIntervalEXT())
        return False

    def set_vsync(self, vsync: bool) -> None:
        if wgl_info.have_extension('WGL_EXT_swap_control'):
            wglext_arb.wglSwapIntervalEXT(int(vsync))


class Win32Context(_BaseWin32Context):  # noqa: D101
    config: Win32DisplayConfig
    context_share: Win32Context | None

    def attach(self, canvas: Win32Canvas) -> None:
        super().attach(canvas)

        if not self._context:
            self.config._set_pixel_format(canvas)  # noqa: SLF001
            self._context = wgl.wglCreateContext(canvas.hdc)

        share = self.context_share
        if share:
            if not share.canvas:
                msg = 'Share context has no canvas.'
                raise RuntimeError(msg)
            if not wgl.wglShareLists(share._context, self._context):  # noqa: SLF001
                msg = 'Unable to share contexts.'
                raise gl.ContextException(msg)


class Win32ARBContext(_BaseWin32Context):  # noqa: D101
    config: Win32DisplayConfigARB
    context_share: Win32ARBContext | None

    def attach(self, canvas: Win32Canvas) -> None:
        share = self.context_share
        if share:
            if not share.canvas:
                msg = 'Share context has no canvas.'
                raise RuntimeError(msg)
            share = share._context  # noqa: SLF001

        attribs = []
        if self.config.major_version is not None:
            attribs.extend([wglext_arb.WGL_CONTEXT_MAJOR_VERSION_ARB, self.config.major_version])
        if self.config.minor_version is not None:
            attribs.extend([wglext_arb.WGL_CONTEXT_MINOR_VERSION_ARB, self.config.minor_version])
        flags = 0
        if self.config.forward_compatible:
            flags |= wglext_arb.WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
        if self.config.debug:
            flags |= wglext_arb.WGL_CONTEXT_DEBUG_BIT_ARB
        if flags:
            attribs.extend([wglext_arb.WGL_CONTEXT_FLAGS_ARB, flags])
        attribs.append(0)
        attribs = (c_int * len(attribs))(*attribs)

        self.config._set_pixel_format(canvas)  # noqa: SLF001
        self._context = wglext_arb.wglCreateContextAttribsARB(canvas.hdc, share, attribs)
        super().attach(canvas)



from ctypes import *

import pyglet.lib


_lib = pyglet.lib.load_library('xkbcommon')


class c_void(Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', c_int)]


class struct_xkb_context(Structure):
    __slots__ = [
    ]


struct_xkb_context._fields_ = [
    ('_opaque_struct', c_int)
]


class struct_xkb_keymap(Structure):
    __slots__ = [
    ]


struct_xkb_keymap._fields_ = [
    ('_opaque_struct', c_int)
]


class struct_xkb_rule_names(Structure):
    _fields_ = [
        ('rules', c_char_p),
        ('model', c_char_p),
        ('layout', c_char_p),
        ('variant', c_char_p),
        ('options', c_char_p),
    ]


class struct_xkb_state(Structure):
    __slots__ = [
    ]


struct_xkb_state._fields_ = [
    ('_opaque_struct', c_int)
]


class struct__IO_FILE(Structure):
    __slots__ = [
    ]


struct__IO_FILE._fields_ = [
    ('_opaque_struct', c_int)
]

XKB_CONTEXT_NO_FLAGS = 0
XKB_CONTEXT_NO_DEFAULT_INCLUDES = (1 << 0)
XKB_CONTEXT_NO_ENVIRONMENT_NAMES = (1 << 1)

XKB_KEYMAP_COMPILE_NO_FLAGS = 0
XKB_KEYMAP_FORMAT_TEXT_V1 = 1

xkb_keycode_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:165
xkb_keysym_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:195
xkb_layout_index_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:224
xkb_layout_mask_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:226
xkb_level_index_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:240
xkb_mod_index_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:265
xkb_mod_mask_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:267
xkb_led_index_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:294
xkb_led_mask_t = c_uint32  # /usr/include/xkbcommon/xkbcommon.h:296
XKB_KEYCODE_INVALID = 4294967295  # /usr/include/xkbcommon/xkbcommon.h:298
XKB_LAYOUT_INVALID = 4294967295  # /usr/include/xkbcommon/xkbcommon.h:299
XKB_LEVEL_INVALID = 4294967295  # /usr/include/xkbcommon/xkbcommon.h:300
XKB_MOD_INVALID = 4294967295  # /usr/include/xkbcommon/xkbcommon.h:301
XKB_LED_INVALID = 4294967295  # /usr/include/xkbcommon/xkbcommon.h:302
XKB_KEYCODE_MAX = 4294967294  # /usr/include/xkbcommon/xkbcommon.h:304
# /usr/include/xkbcommon/xkbcommon.h:438
xkb_keysym_get_name = _lib.xkb_keysym_get_name
xkb_keysym_get_name.restype = c_int
xkb_keysym_get_name.argtypes = [xkb_keysym_t, c_char_p, c_size_t]

enum_xkb_keysym_flags = c_int
# /usr/include/xkbcommon/xkbcommon.h:472
xkb_keysym_from_name = _lib.xkb_keysym_from_name
xkb_keysym_from_name.restype = xkb_keysym_t
xkb_keysym_from_name.argtypes = [c_char_p, enum_xkb_keysym_flags]

# /usr/include/xkbcommon/xkbcommon.h:491
xkb_keysym_to_utf8 = _lib.xkb_keysym_to_utf8
xkb_keysym_to_utf8.restype = c_int
xkb_keysym_to_utf8.argtypes = [xkb_keysym_t, c_char_p, c_size_t]

# /usr/include/xkbcommon/xkbcommon.h:506
xkb_keysym_to_utf32 = _lib.xkb_keysym_to_utf32
xkb_keysym_to_utf32.restype = c_uint32
xkb_keysym_to_utf32.argtypes = [xkb_keysym_t]

# /usr/include/xkbcommon/xkbcommon.h:529
xkb_utf32_to_keysym = _lib.xkb_utf32_to_keysym
xkb_utf32_to_keysym.restype = xkb_keysym_t
xkb_utf32_to_keysym.argtypes = [c_uint32]

# /usr/include/xkbcommon/xkbcommon.h:540
xkb_keysym_to_upper = _lib.xkb_keysym_to_upper
xkb_keysym_to_upper.restype = xkb_keysym_t
xkb_keysym_to_upper.argtypes = [xkb_keysym_t]

# /usr/include/xkbcommon/xkbcommon.h:549
xkb_keysym_to_lower = _lib.xkb_keysym_to_lower
xkb_keysym_to_lower.restype = xkb_keysym_t
xkb_keysym_to_lower.argtypes = [xkb_keysym_t]

enum_xkb_context_flags = c_int
# /usr/include/xkbcommon/xkbcommon.h:597
xkb_context_new = _lib.xkb_context_new
xkb_context_new.restype = POINTER(struct_xkb_context)
xkb_context_new.argtypes = [enum_xkb_context_flags]

# /usr/include/xkbcommon/xkbcommon.h:607
xkb_context_ref = _lib.xkb_context_ref
xkb_context_ref.restype = POINTER(struct_xkb_context)
xkb_context_ref.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:618
xkb_context_unref = _lib.xkb_context_unref
xkb_context_unref.restype = None
xkb_context_unref.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:629
xkb_context_set_user_data = _lib.xkb_context_set_user_data
xkb_context_set_user_data.restype = None
xkb_context_set_user_data.argtypes = [POINTER(struct_xkb_context), POINTER(None)]

# /usr/include/xkbcommon/xkbcommon.h:642
xkb_context_get_user_data = _lib.xkb_context_get_user_data
xkb_context_get_user_data.restype = POINTER(c_void)
xkb_context_get_user_data.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:677
xkb_context_include_path_append = _lib.xkb_context_include_path_append
xkb_context_include_path_append.restype = c_int
xkb_context_include_path_append.argtypes = [POINTER(struct_xkb_context), c_char_p]

# /usr/include/xkbcommon/xkbcommon.h:687
xkb_context_include_path_append_default = _lib.xkb_context_include_path_append_default
xkb_context_include_path_append_default.restype = c_int
xkb_context_include_path_append_default.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:700
xkb_context_include_path_reset_defaults = _lib.xkb_context_include_path_reset_defaults
xkb_context_include_path_reset_defaults.restype = c_int
xkb_context_include_path_reset_defaults.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:708
xkb_context_include_path_clear = _lib.xkb_context_include_path_clear
xkb_context_include_path_clear.restype = None
xkb_context_include_path_clear.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:716
xkb_context_num_include_paths = _lib.xkb_context_num_include_paths
xkb_context_num_include_paths.restype = c_uint
xkb_context_num_include_paths.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:726
xkb_context_include_path_get = _lib.xkb_context_include_path_get
xkb_context_include_path_get.restype = c_char_p
xkb_context_include_path_get.argtypes = [POINTER(struct_xkb_context), c_uint]

enum_xkb_log_level = c_int
# /usr/include/xkbcommon/xkbcommon.h:761
xkb_context_set_log_level = _lib.xkb_context_set_log_level
xkb_context_set_log_level.restype = None
xkb_context_set_log_level.argtypes = [POINTER(struct_xkb_context), enum_xkb_log_level]

# /usr/include/xkbcommon/xkbcommon.h:770
xkb_context_get_log_level = _lib.xkb_context_get_log_level
xkb_context_get_log_level.restype = enum_xkb_log_level
xkb_context_get_log_level.argtypes = [POINTER(struct_xkb_context)]

# /usr/include/xkbcommon/xkbcommon.h:792
xkb_context_set_log_verbosity = _lib.xkb_context_set_log_verbosity
xkb_context_set_log_verbosity.restype = None
xkb_context_set_log_verbosity.argtypes = [POINTER(struct_xkb_context), c_int]

# /usr/include/xkbcommon/xkbcommon.h:800
xkb_context_get_log_verbosity = _lib.xkb_context_get_log_verbosity
xkb_context_get_log_verbosity.restype = c_int
xkb_context_get_log_verbosity.argtypes = [POINTER(struct_xkb_context)]

enum_xkb_keymap_compile_flags = c_int
# /usr/include/xkbcommon/xkbcommon.h:859
xkb_keymap_new_from_names = _lib.xkb_keymap_new_from_names
xkb_keymap_new_from_names.restype = POINTER(struct_xkb_keymap)
xkb_keymap_new_from_names.argtypes = [POINTER(struct_xkb_context), POINTER(struct_xkb_rule_names),
                                      enum_xkb_keymap_compile_flags]

FILE = struct__IO_FILE  # /usr/include/bits/types/FILE.h:7
enum_xkb_keymap_format = c_int
# /usr/include/xkbcommon/xkbcommon.h:888
xkb_keymap_new_from_file = _lib.xkb_keymap_new_from_file
xkb_keymap_new_from_file.restype = POINTER(struct_xkb_keymap)
xkb_keymap_new_from_file.argtypes = [POINTER(struct_xkb_context), POINTER(FILE), enum_xkb_keymap_format,
                                     enum_xkb_keymap_compile_flags]

# /usr/include/xkbcommon/xkbcommon.h:902
xkb_keymap_new_from_string = _lib.xkb_keymap_new_from_string
xkb_keymap_new_from_string.restype = POINTER(struct_xkb_keymap)
xkb_keymap_new_from_string.argtypes = [POINTER(struct_xkb_context), c_char_p, enum_xkb_keymap_format,
                                       enum_xkb_keymap_compile_flags]

# /usr/include/xkbcommon/xkbcommon.h:917
xkb_keymap_new_from_buffer = _lib.xkb_keymap_new_from_buffer
xkb_keymap_new_from_buffer.restype = POINTER(struct_xkb_keymap)
xkb_keymap_new_from_buffer.argtypes = [POINTER(struct_xkb_context), c_char_p, c_size_t, enum_xkb_keymap_format,
                                       enum_xkb_keymap_compile_flags]

# /usr/include/xkbcommon/xkbcommon.h:929
xkb_keymap_ref = _lib.xkb_keymap_ref
xkb_keymap_ref.restype = POINTER(struct_xkb_keymap)
xkb_keymap_ref.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:940
xkb_keymap_unref = _lib.xkb_keymap_unref
xkb_keymap_unref.restype = None
xkb_keymap_unref.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:966
xkb_keymap_get_as_string = _lib.xkb_keymap_get_as_string
xkb_keymap_get_as_string.restype = c_char_p
xkb_keymap_get_as_string.argtypes = [POINTER(struct_xkb_keymap), enum_xkb_keymap_format]

# /usr/include/xkbcommon/xkbcommon.h:987
xkb_keymap_min_keycode = _lib.xkb_keymap_min_keycode
xkb_keymap_min_keycode.restype = xkb_keycode_t
xkb_keymap_min_keycode.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:997
xkb_keymap_max_keycode = _lib.xkb_keymap_max_keycode
xkb_keymap_max_keycode.restype = xkb_keycode_t
xkb_keymap_max_keycode.argtypes = [POINTER(struct_xkb_keymap)]

xkb_keymap_key_iter_t = CFUNCTYPE(None, POINTER(struct_xkb_keymap), xkb_keycode_t, POINTER(None))

# /usr/include/xkbcommon/xkbcommon.h:1020
xkb_keymap_key_for_each = _lib.xkb_keymap_key_for_each
xkb_keymap_key_for_each.restype = None
xkb_keymap_key_for_each.argtypes = [POINTER(struct_xkb_keymap), xkb_keymap_key_iter_t, POINTER(None)]

# /usr/include/xkbcommon/xkbcommon.h:1036
xkb_keymap_key_get_name = _lib.xkb_keymap_key_get_name
xkb_keymap_key_get_name.restype = c_char_p
xkb_keymap_key_get_name.argtypes = [POINTER(struct_xkb_keymap), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1052
xkb_keymap_key_by_name = _lib.xkb_keymap_key_by_name
xkb_keymap_key_by_name.restype = xkb_keycode_t
xkb_keymap_key_by_name.argtypes = [POINTER(struct_xkb_keymap), c_char_p]

# /usr/include/xkbcommon/xkbcommon.h:1061
xkb_keymap_num_mods = _lib.xkb_keymap_num_mods
xkb_keymap_num_mods.restype = xkb_mod_index_t
xkb_keymap_num_mods.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:1071
xkb_keymap_mod_get_name = _lib.xkb_keymap_mod_get_name
xkb_keymap_mod_get_name.restype = c_char_p
xkb_keymap_mod_get_name.argtypes = [POINTER(struct_xkb_keymap), xkb_mod_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1084
xkb_keymap_mod_get_index = _lib.xkb_keymap_mod_get_index
xkb_keymap_mod_get_index.restype = xkb_mod_index_t
xkb_keymap_mod_get_index.argtypes = [POINTER(struct_xkb_keymap), c_char_p]

# /usr/include/xkbcommon/xkbcommon.h:1093
xkb_keymap_num_layouts = _lib.xkb_keymap_num_layouts
xkb_keymap_num_layouts.restype = xkb_layout_index_t
xkb_keymap_num_layouts.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:1105
xkb_keymap_layout_get_name = _lib.xkb_keymap_layout_get_name
xkb_keymap_layout_get_name.restype = c_char_p
xkb_keymap_layout_get_name.argtypes = [POINTER(struct_xkb_keymap), xkb_layout_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1120
xkb_keymap_layout_get_index = _lib.xkb_keymap_layout_get_index
xkb_keymap_layout_get_index.restype = xkb_layout_index_t
xkb_keymap_layout_get_index.argtypes = [POINTER(struct_xkb_keymap), c_char_p]

# /usr/include/xkbcommon/xkbcommon.h:1134
xkb_keymap_num_leds = _lib.xkb_keymap_num_leds
xkb_keymap_num_leds.restype = xkb_led_index_t
xkb_keymap_num_leds.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:1143
xkb_keymap_led_get_name = _lib.xkb_keymap_led_get_name
xkb_keymap_led_get_name.restype = c_char_p
xkb_keymap_led_get_name.argtypes = [POINTER(struct_xkb_keymap), xkb_led_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1155
xkb_keymap_led_get_index = _lib.xkb_keymap_led_get_index
xkb_keymap_led_get_index.restype = xkb_led_index_t
xkb_keymap_led_get_index.argtypes = [POINTER(struct_xkb_keymap), c_char_p]

# /usr/include/xkbcommon/xkbcommon.h:1168
xkb_keymap_num_layouts_for_key = _lib.xkb_keymap_num_layouts_for_key
xkb_keymap_num_layouts_for_key.restype = xkb_layout_index_t
xkb_keymap_num_layouts_for_key.argtypes = [POINTER(struct_xkb_keymap), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1181
xkb_keymap_num_levels_for_key = _lib.xkb_keymap_num_levels_for_key
xkb_keymap_num_levels_for_key.restype = xkb_level_index_t
xkb_keymap_num_levels_for_key.argtypes = [POINTER(struct_xkb_keymap), xkb_keycode_t, xkb_layout_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1221
xkb_keymap_key_get_mods_for_level = _lib.xkb_keymap_key_get_mods_for_level
xkb_keymap_key_get_mods_for_level.restype = c_size_t
xkb_keymap_key_get_mods_for_level.argtypes = [POINTER(struct_xkb_keymap), xkb_keycode_t, xkb_layout_index_t,
                                              xkb_level_index_t, POINTER(xkb_mod_mask_t), c_size_t]

# /usr/include/xkbcommon/xkbcommon.h:1257
xkb_keymap_key_get_syms_by_level = _lib.xkb_keymap_key_get_syms_by_level
xkb_keymap_key_get_syms_by_level.restype = c_int
xkb_keymap_key_get_syms_by_level.argtypes = [POINTER(struct_xkb_keymap), xkb_keycode_t, xkb_layout_index_t,
                                             xkb_level_index_t, POINTER(POINTER(xkb_keysym_t))]

# /usr/include/xkbcommon/xkbcommon.h:1279
xkb_keymap_key_repeats = _lib.xkb_keymap_key_repeats
xkb_keymap_key_repeats.restype = c_int
xkb_keymap_key_repeats.argtypes = [POINTER(struct_xkb_keymap), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1299
xkb_state_new = _lib.xkb_state_new
xkb_state_new.restype = POINTER(struct_xkb_state)
xkb_state_new.argtypes = [POINTER(struct_xkb_keymap)]

# /usr/include/xkbcommon/xkbcommon.h:1309
xkb_state_ref = _lib.xkb_state_ref
xkb_state_ref.restype = POINTER(struct_xkb_state)
xkb_state_ref.argtypes = [POINTER(struct_xkb_state)]

# /usr/include/xkbcommon/xkbcommon.h:1320
xkb_state_unref = _lib.xkb_state_unref
xkb_state_unref.restype = None
xkb_state_unref.argtypes = [POINTER(struct_xkb_state)]

# /usr/include/xkbcommon/xkbcommon.h:1334
xkb_state_get_keymap = _lib.xkb_state_get_keymap
xkb_state_get_keymap.restype = POINTER(struct_xkb_keymap)
xkb_state_get_keymap.argtypes = [POINTER(struct_xkb_state)]

enum_xkb_state_component = c_int
enum_xkb_key_direction = c_int

# /usr/include/xkbcommon/xkbcommon.h:1409
xkb_state_update_key = _lib.xkb_state_update_key
xkb_state_update_key.restype = enum_xkb_state_component
xkb_state_update_key.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, enum_xkb_key_direction]

# /usr/include/xkbcommon/xkbcommon.h:1440
xkb_state_update_mask = _lib.xkb_state_update_mask
xkb_state_update_mask.restype = enum_xkb_state_component
xkb_state_update_mask.argtypes = [POINTER(struct_xkb_state), xkb_mod_mask_t, xkb_mod_mask_t, xkb_mod_mask_t,
                                  xkb_layout_index_t, xkb_layout_index_t, xkb_layout_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1475
xkb_state_key_get_syms = _lib.xkb_state_key_get_syms
xkb_state_key_get_syms.restype = c_int
xkb_state_key_get_syms.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, POINTER(POINTER(xkb_keysym_t))]

# /usr/include/xkbcommon/xkbcommon.h:1505
xkb_state_key_get_utf8 = _lib.xkb_state_key_get_utf8
xkb_state_key_get_utf8.restype = c_int
xkb_state_key_get_utf8.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, c_char_p, c_size_t]

# /usr/include/xkbcommon/xkbcommon.h:1522
xkb_state_key_get_utf32 = _lib.xkb_state_key_get_utf32
xkb_state_key_get_utf32.restype = c_uint32
xkb_state_key_get_utf32.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1542
xkb_state_key_get_one_sym = _lib.xkb_state_key_get_one_sym
xkb_state_key_get_one_sym.restype = xkb_keysym_t
xkb_state_key_get_one_sym.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1559
xkb_state_key_get_layout = _lib.xkb_state_key_get_layout
xkb_state_key_get_layout.restype = xkb_layout_index_t
xkb_state_key_get_layout.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1584
xkb_state_key_get_level = _lib.xkb_state_key_get_level
xkb_state_key_get_level.restype = xkb_level_index_t
xkb_state_key_get_level.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, xkb_layout_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1622
xkb_state_serialize_mods = _lib.xkb_state_serialize_mods
xkb_state_serialize_mods.restype = xkb_mod_mask_t
xkb_state_serialize_mods.argtypes = [POINTER(struct_xkb_state), enum_xkb_state_component]

# /usr/include/xkbcommon/xkbcommon.h:1644
xkb_state_serialize_layout = _lib.xkb_state_serialize_layout
xkb_state_serialize_layout.restype = xkb_layout_index_t
xkb_state_serialize_layout.argtypes = [POINTER(struct_xkb_state), enum_xkb_state_component]

# /usr/include/xkbcommon/xkbcommon.h:1656
xkb_state_mod_name_is_active = _lib.xkb_state_mod_name_is_active
xkb_state_mod_name_is_active.restype = c_int
xkb_state_mod_name_is_active.argtypes = [POINTER(struct_xkb_state), c_char_p, enum_xkb_state_component]

enum_xkb_state_match = c_int
# /usr/include/xkbcommon/xkbcommon.h:1677
xkb_state_mod_names_are_active = _lib.xkb_state_mod_names_are_active
xkb_state_mod_names_are_active.restype = c_int
xkb_state_mod_names_are_active.argtypes = [POINTER(struct_xkb_state), enum_xkb_state_component, enum_xkb_state_match]

# /usr/include/xkbcommon/xkbcommon.h:1691
xkb_state_mod_index_is_active = _lib.xkb_state_mod_index_is_active
xkb_state_mod_index_is_active.restype = c_int
xkb_state_mod_index_is_active.argtypes = [POINTER(struct_xkb_state), xkb_mod_index_t, enum_xkb_state_component]

# /usr/include/xkbcommon/xkbcommon.h:1712
xkb_state_mod_indices_are_active = _lib.xkb_state_mod_indices_are_active
xkb_state_mod_indices_are_active.restype = c_int
xkb_state_mod_indices_are_active.argtypes = [POINTER(struct_xkb_state), enum_xkb_state_component, enum_xkb_state_match]

enum_xkb_consumed_mode = c_int
# /usr/include/xkbcommon/xkbcommon.h:1832
xkb_state_key_get_consumed_mods2 = _lib.xkb_state_key_get_consumed_mods2
xkb_state_key_get_consumed_mods2.restype = xkb_mod_mask_t
xkb_state_key_get_consumed_mods2.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, enum_xkb_consumed_mode]

# /usr/include/xkbcommon/xkbcommon.h:1842
xkb_state_key_get_consumed_mods = _lib.xkb_state_key_get_consumed_mods
xkb_state_key_get_consumed_mods.restype = xkb_mod_mask_t
xkb_state_key_get_consumed_mods.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t]

# /usr/include/xkbcommon/xkbcommon.h:1862
xkb_state_mod_index_is_consumed2 = _lib.xkb_state_mod_index_is_consumed2
xkb_state_mod_index_is_consumed2.restype = c_int
xkb_state_mod_index_is_consumed2.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, xkb_mod_index_t,
                                             enum_xkb_consumed_mode]

# /usr/include/xkbcommon/xkbcommon.h:1874
xkb_state_mod_index_is_consumed = _lib.xkb_state_mod_index_is_consumed
xkb_state_mod_index_is_consumed.restype = c_int
xkb_state_mod_index_is_consumed.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, xkb_mod_index_t]

# /usr/include/xkbcommon/xkbcommon.h:1889
xkb_state_mod_mask_remove_consumed = _lib.xkb_state_mod_mask_remove_consumed
xkb_state_mod_mask_remove_consumed.restype = xkb_mod_mask_t
xkb_state_mod_mask_remove_consumed.argtypes = [POINTER(struct_xkb_state), xkb_keycode_t, xkb_mod_mask_t]

# /usr/include/xkbcommon/xkbcommon.h:1905
xkb_state_layout_name_is_active = _lib.xkb_state_layout_name_is_active
xkb_state_layout_name_is_active.restype = c_int
xkb_state_layout_name_is_active.argtypes = [POINTER(struct_xkb_state), c_char_p, enum_xkb_state_component]

# /usr/include/xkbcommon/xkbcommon.h:1918
xkb_state_layout_index_is_active = _lib.xkb_state_layout_index_is_active
xkb_state_layout_index_is_active.restype = c_int
xkb_state_layout_index_is_active.argtypes = [POINTER(struct_xkb_state), xkb_layout_index_t, enum_xkb_state_component]

# /usr/include/xkbcommon/xkbcommon.h:1932
xkb_state_led_name_is_active = _lib.xkb_state_led_name_is_active
xkb_state_led_name_is_active.restype = c_int
xkb_state_led_name_is_active.argtypes = [POINTER(struct_xkb_state), c_char_p]

# /usr/include/xkbcommon/xkbcommon.h:1944
xkb_state_led_index_is_active = _lib.xkb_state_led_index_is_active
xkb_state_led_index_is_active.restype = c_int
xkb_state_led_index_is_active.argtypes = [POINTER(struct_xkb_state), xkb_led_index_t]



"""configure.py script to get / set some session configurable values; mp.py is
an alias for this command.
"""

import sys

import fs
import mpexceptions


template_cmd = (
"""
Usage

    &cmd subcommand [args]

Subcommands

    new session [playlist] : Creates a new session, sets it as the active one
    activate session : activates a session 
    deactivate : no session will be active
    protect [target]: forbids overwrite of session data
    status : prints configuration for the active session
    help [subcommand] : prints help for the given subcommand or topic
    list : list all sessions associated the current samples_dir 

Creates and manages pyglet media_player debug session configurations.

Most commands and subcommands need an environment variable pyglet_mp_samples_dir
to be set to the directory where the media samples reside.

The configuration stores some values used when other commands are executed.

    samples_dir: directory where the samples reside
    session: a name to identify a particular session
    session_dir: directory to store all the session's output files, set to  
                 samples_dir/session
    permissions to overwrite or not raw data collected or generated reports
    playlist: subset of samples to play

This command can be called both as 'configure.py' or 'mp.py', they do the same.

"""
)

template_subcommands = (
"""
new
Usage

    &cmd new session [playlist_file]

Arguments

    session       : a name to identify this particular session
    playlist_file : file with a list of samples

Creates a new session and a directory to save session outputs.

The configuration is saved to disk and the session is made the active one.

If playlist_file os specified, only the samples in the playlist will be
considered as targets for a general command. 

General commands will use the configuration of the active session to know from
where to read and write data.

@
activate
Usage

    &cmd activate session

Arguments

    session: name of session to activate.

Sets the session as the active one.

General commands will use the configuration of the active session to know from
where to read and write data.

@
deactivate
Usage

    &cmd deactivate

Makes the current session inactive.

No session will be active, so general commands will not modify any session data.

Mostly as a precaution so that a slip in shell history does not overwrite
anything.
@
protect
Usage

    &cmd protect target

Arguments

target: one of

"raw_data"
    Forbids general commands to overwrite session's raw data captured.
    Helps to not mix in the same session results obtained in different conditions.

"reports"
    Forbids general commands to overwrite session's generated reports.
    Useful if the reports are manually annotated.

@
status
Usage

    &cmd status [session]

Arguments

    session: name of session

Prints the configuration for the active session.

@
help
Usage

    &cmd help [subcommand or topic]

Subcommands

    new session [playlist] : Creates a new session, sets it as the active one
    activate session : activates a session 
    deactivate : no session will be active
    protect [target]: forbids overwrite of session data
    status : prints configuration for the active session
    help [subcommand] : prints help for the given subcommand or topic
    list : list all sessions associated the current samples_dir 

Topics

    layout  : data layout on disk
    session : sessions what and whys
    workflow: common workflows
    gencmds : general commands
@
list
Usage

    &cmd list

List all sessions associated with the current sample_dir


"""
)


def sub_and_template_from_section(s):
    subcommand, template = s[1:].split("\n", 1)
    return subcommand, template


def help_texts_from_template(cmd):
    cmd_help = template_cmd.replace("&cmd", cmd)

    all_subcommands = template_subcommands.replace("&cmd", cmd)
    parts = all_subcommands.split("@")
    pairs = [sub_and_template_from_section(s) for s in parts]
    subcommands_help = {a: b for a, b in pairs}

    return cmd_help, subcommands_help


def test_help_strings():
    cmd_help, subcommands_help = help_texts_from_template('mp')
    print(cmd_help)
    for e in subcommands_help:
        print("sub:", repr(e))
        print("|%s|" % subcommands_help[e])


def show_configuration(keys_to_show=None):
    pathserv = fs.get_path_info_for_active_session()
    conf = fs.json_load(pathserv.configuration_filename())
    if keys_to_show is None:
        keys_to_show = conf.keys()
    print("session:", pathserv.session)
    for k in sorted(keys_to_show):
        print("\t%s: %s" % (k, conf[k]))


def sub_activate():
    if len(sys.argv) < 3:
        print("*** Error, missing argument.\n")
        print(subcommands_help["activate"])
        return 1
    session = sys.argv[2]
    fs.activation_set_to(session)
    return 0


def sub_deactivate():
    session = None
    fs.activation_set_to(session)
    return 0


def sub_help():
    if len(sys.argv) < 3:
        topic = "help"
    else:
        topic = sys.argv[2]
    print(subcommands_help[topic])
    return 0


def sub_list():
    samples_dir = fs.get_current_samples_dir()
    sessions = fs.get_sessions(samples_dir)
    for session in sorted(sessions):
        print(session)
    return 0


def sub_new():
    if len(sys.argv) < 3:
        print("*** Error, missing argument.\n")
        print(subcommands_help["new"])
        return 1

    session = sys.argv[2]
    if len(sys.argv) > 3:
        playlist_file = sys.argv[3]
    else:
        playlist_file = None
    fs.new_session(session, playlist_file)
    return 0


def sub_protect():
    if len(sys.argv) < 3:
        print("*** Error, missing argument.\n")
        print(subcommands_help["protect"])
        return 1

    name = sys.argv[2]
    if name not in {"raw_data", "reports"}:
        print("*** Error, unknown name to protect. name:", name)
        print(subcommands_help["protect"])
        return 1
    modify = {("protect_" + name): True}
    fs.update_active_configuration(modify)
    return 0


def sub_status():
    show_configuration()
    return 0


def dispatch_subcommand(caller):
    global cmd_help, subcommands_help
    cmd_help, subcommands_help = help_texts_from_template(caller)
    if len(sys.argv) < 2:
        sub = "help"
    else:
        sub = sys.argv[1]
    try:
        returncode = globals()["sub_" + sub]()
    except KeyError:
        print("*** Error, unknown subcommand:", sub)
        returncode = 1
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined")
        returncode = 1
    except mpexceptions.ExceptionSamplesDirDoesNotExist as ex:
        print("The env var 'pyglet_mp_samples_dir' does not point to an existing directory")
        returncode = 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session is active.")
        returncode = 1
    except mpexceptions.ExceptionNoSessionWithThatName:
        print("*** Error, no session by that name")
        returncode = 1
    except mpexceptions.ExceptionSessionExistWithSameName:
        print("*** Error, session exists")
        returncode = 1
    except mpexceptions.ExceptionPlaylistFileDoNotExists:
        print("*** Error, playlist_file does not exist")
        returncode = 1
    except mpexceptions.ExceptionBadSampleInPlaylist as ex:
        print("*** Error, bad sample(s) name in playlist (bad extension or non-existent):")
        for sample in ex.rejected:
            print("\t%s" % sample)
        returncode = 1

    sys.exit(returncode)

if __name__ == "__main__":
    dispatch_subcommand("configure.py")



"""Matrix and Vector math.

This module provides Vector and Matrix objects, including Vec2, Vec3,
Vec4, Mat3, and Mat4. Most common matrix and vector operations are
supported. Helper methods are included for rotating, scaling, and
transforming. The :py:class:`~pyglet.matrix.Mat4` includes class methods
for creating orthographic and perspective projection matrixes.

Matrices behave just like they do in GLSL: they are specified in column-major
order and multiply on the left of vectors, which are treated as columns.

All objects are immutable and hashable.
"""
# WARNING! DO NOT TRY TO MAKE THIS FILE "PRETTIER"
#
# The unusual code style in this module is an intentional sacrifice:
# * It allows decent performance despite being written in pure Python
# * It ensures good developer experience outside pyglet.math, especially
#   on systems without access to compute shaders
from __future__ import annotations

import math as _math
import typing as _typing
import warnings as _warnings


def clamp(num: float, minimum: float, maximum: float) -> float:
    """Clamp a value between a minimum and maximum limit."""
    return max(min(num, maximum), minimum)


class Vec2(_typing.NamedTuple):
    """A two-dimensional vector represented as an X Y coordinate pair.

    `Vec2` is an immutable 2D Vector, including most common
    operators. As an immutable type, all operations return a new object.

    .. note:: The Python ``len`` operator returns the number of elements in
              the vector. For the vector length, use the `length()` method.

    .. note:: Python's :py:func:`sum` requires the first item to be a ``Vec2``.

              After that, you can mix ``Vec2``-like :py:class:`tuple`
              and ``Vec2`` instances freely in the iterable.

              If you do not, you will see a :py:class:`TypeError` about being
              unable to add a :py:class:`tuple` and a :py:class:`int`.

    """

    x: float = 0.0
    y: float = 0.0

    def __bool__(self):
        return self.x != 0.0 or self.y != 0.0

    def __add__(self, other: Vec2 | tuple[float, float] | float) -> Vec2:
        try:
            return Vec2(
                self[0] + other[0], self[1] + other[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] + other, self[1] + other  # type: ignore
            )

    def __radd__(self, other: Vec2 | tuple[float, float] | float) -> Vec2:
        try:
            return self.__add__(other)
        except TypeError as err:
            if other == 0:  # Required for functionality with sum()
                return self
            raise err

    def __sub__(self, other: Vec2 | tuple[float, float] | float) -> Vec2:
        try:
            return Vec2(
                self[0] - other[0], self[1] - other[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] - other, self[1] - other  # type: ignore
            )

    def __rsub__(self, other: Vec2 | tuple[float, float] | float) -> Vec2:
        try:
            return Vec2(
                other[0] - self[0], other[1] - self[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                other - self[0], other - self[1]  # type: ignore
            )

    def __mul__(self, scalar: float | Vec2 | tuple[float, float]) -> Vec2:
        try:
            return Vec2(
                self[0] * scalar[0], self[1] * scalar[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] * scalar, self[1] * scalar  # type: ignore
            )

    def __rmul__(self, scalar: float | Vec2 | tuple[float, float]) -> Vec2:
        try:
            return Vec2(
                self[0] * scalar[0], self[1] * scalar[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] * scalar, self[1] * scalar  # type: ignore
            )

    def __truediv__(self, scalar: float | Vec2 | tuple[float, float]) -> Vec2:
        try:
            return Vec2(
                self[0] / scalar[0], self[1] / scalar[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] / scalar, self[1] / scalar  # type: ignore
            )

    def __rtruediv__(self, scalar: float | Vec2 | tuple[float, float]) -> Vec2:
        try:
            return Vec2(
                scalar[0] / self[0], scalar[1] / self[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                scalar / self[0], scalar / self[1]  # type: ignore
            )

    def __floordiv__(self, scalar: float | Vec2 | tuple[float, float]) -> Vec2:
        try:
            return Vec2(
                self.x // scalar[0], self.y // scalar[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] // scalar, self[1] // scalar  # type: ignore
            )

    def __rfloordiv__(self, scalar: float | Vec2 | tuple[float, float]) -> Vec2:
        try:
            return Vec2(
                scalar[0] // self[0], scalar[1] // self[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                scalar // self[0], scalar // self[1]  # type: ignore
            )

    def __abs__(self) -> Vec2:
        return Vec2(abs(self[0]), abs(self[1]))

    def __neg__(self) -> Vec2:
        return Vec2(-self[0], -self[1])

    def __round__(self, n_digits: int | None = None) -> Vec2:
        return Vec2(*(round(v, n_digits) for v in self))

    def __ceil__(self) -> Vec2:
        return Vec2(_math.ceil(self[0]), _math.ceil(self[1]))

    def __floor__(self) -> Vec2:
        return Vec2(_math.floor(self[0]), _math.floor(self[1]))

    def __trunc__(self) -> Vec2:
        return Vec2(_math.trunc(self[0]), _math.trunc(self[1]))

    def __mod__(self, other: Vec2 | tuple[float, float] | float) -> Vec2:
        try:
            return Vec2(
                self[0] % other[0], self[1] % other[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] % other, self[1] % other  # type: ignore
            )

    def __pow__(self, other: Vec2 | tuple[float, float] | float) -> Vec2:
        try:
            return Vec2(
                self[0] ** other[0], self[1] ** other[1]  # type: ignore
            )
        except TypeError:
            return Vec2(
                self[0] ** other, self[1] ** other  # type: ignore
            )

    def __lt__(self, other: tuple[float, float]) -> bool:
        return self[0] ** 2 + self[1] ** 2 < other[0] ** 2 + other[1] ** 2

    @staticmethod
    def from_heading(heading: float, length: float = 1.0) -> Vec2:
        """Create a new vector from the given heading and length.

        Args:
          heading: The desired heading, in radians
          length: The desired length of the vector
        """
        return Vec2(length * _math.cos(heading), length * _math.sin(heading))

    @staticmethod
    def from_polar(angle: float, length: float = 1.0) -> Vec2:
        """Create a new vector from the given polar coordinates.

        Args:
          angle: The angle, in radians.
          length: The desired length
        """
        return Vec2(length * _math.cos(angle), length * _math.sin(angle))

    def length(self) -> float:
        """Calculate the length of the vector: ``sqrt(x ** 2 + y ** 2)``."""
        return _math.sqrt(self[0] ** 2 + self[1] ** 2)

    def heading(self) -> float:
        """Calculate the heading of the vector in radians.

        Shortcut for `atan2(y, x)` meaning it returns a value between
        -pi and pi. ``Vec2(1, 0)`` will have a heading of 0. Counter-clockwise
        is positive moving towards pi, and clockwise is negative moving towards -pi.
        """
        return _math.atan2(self[1], self[0])

    def length_squared(self) -> float:
        """Calculate the squared length of the vector.

        This is simply shortcut for `x ** 2 + y ** 2` and can be used
        for faster comparisons without the need for a square root.
        """
        return self[0] ** 2.0 + self[1] ** 2.0

    def lerp(self, other: Vec2 | tuple[float, float], amount: float) -> Vec2:
        """Create a new Vec2 linearly interpolated between this vector and another Vec2.

        The equivalent in GLSL is `mix`.

        Args:
          other: Another Vec2 instance.
          amount: The amount of interpolation between this vector, and the other
                  vector. This should be a value between 0.0 and 1.0. For example:
                  0.5 is the midway point between both vectors.
        """
        return Vec2(self[0] + (amount * (other[0] - self.x)), self[1] + (amount * (other[1] - self[1])))

    def step(self, edge: Vec2 | tuple[float, float]) -> Vec2:
        """A step function that returns 0.0 for a component if it is less than the edge, and 1.0 otherwise.

        This can be used enable and disable some behavior based on a condition.

        Example::

            # First component is less than 1.0, second component is greater than 1.0
            Vec2(0.5, 1.5).step((1.0, 1.0))
            Vec2(1.0, 0.0)

        Args:
            edge: A Vec2 instance.
        """
        return Vec2(0.0 if self[0] < edge[0] else 1.0, 0.0 if self[1] < edge[1] else 1.0)

    def reflect(self, vector: Vec2 | tuple[float, float]) -> Vec2:
        """Create a new Vec2 reflected (ricochet) from the given normalized vector.

        Args:
            vector: A normalized Vec2 or Vec2-like tuple.
        """
        # Less unrolled equivalent of code below:
        #  return self - vector * 2 * vector.dot(self)
        twice_dot_value = self.dot(vector) * 2
        return Vec2(
            self[0] - twice_dot_value * vector[0],
            self[1] - twice_dot_value * vector[1]
        )

    def rotate(self, angle: float) -> Vec2:
        """Create a new vector rotated by the angle. The length remains unchanged.

        Args:
            angle: The desired angle, in radians.
        """
        s = _math.sin(angle)
        c = _math.cos(angle)
        return Vec2(c * self[0] - s * self[1], s * self[0] + c * self[1])

    def distance(self, other: Vec2 | tuple[int, int]) -> float:
        """Calculate the distance between this vector and another vector.

        Args:
            other: The point to calculate the distance to.
        """
        return _math.sqrt(((other[0] - self[0]) ** 2) + ((other[1] - self[1]) ** 2))

    def normalize(self) -> Vec2:
        """Return a normalized version of the vector.

        This simply means the vector will have a length of 1.0. If the vector
        has a length of 0, the original vector will be returned.
        """
        if d := _math.sqrt(self[0] ** 2 + self[1] ** 2):
            return Vec2(self[0] / d, self[1] / d)
        return self

    if _typing.TYPE_CHECKING:
        @_typing.overload
        def clamp(self, min_val: float, max_val: float) -> Vec2:
            ...

        @_typing.overload
        def clamp(self, min_val: Vec2 | tuple[float, float], max_val: Vec2 | tuple[float, float]) -> Vec2:
            ...

        # -- Begin revert if perf-impacting block --
        @_typing.overload
        def clamp(self, min_val: Vec2 | tuple[float, float], max_val: float) -> Vec2:
            ...

        @_typing.overload
        def clamp(self, min_val: float, max_val: Vec2 | tuple[float, float]) -> Vec2:
            ...
        # -- End revert if perf-impacting block --

    def clamp(self, min_val: Vec2 | tuple[float, float] | float, max_val: Vec2 | tuple[float, float] | float) -> Vec2:
        """Restrict the value of the X and Y components of the vector to be within the given values.

        If a single value is provided, it will be used for both X and Y.
        If a tuple or Vec2 is provided, the first value will be used for X and the second for Y.

        Args:
            min_val: The minimum value(s)
            max_val: The maximum value(s)
        """
        try:
            min_x, min_y = min_val[0], min_val[1]  # type: ignore
        except TypeError:
            min_x = min_val
            min_y = min_val
        try:
            max_x, max_y = max_val[0], max_val[1]  # type: ignore
        except TypeError:
            max_x = max_val
            max_y = max_val

        return Vec2(
            clamp(self[0], min_x, max_x), clamp(self[1], min_y, max_y),  # type: ignore
        )

    def dot(self, other: Vec2 | tuple[float, float]) -> float:
        """Calculate the dot product of this vector and another 2D vector."""
        return self[0] * other[0] + self[1] * other[1]

    def index(self, *args: _typing.Any) -> int:
        raise NotImplementedError("Vec types can be indexed directly.")

    def __getattr__(self, attrs: str) -> Vec2 | Vec3 | Vec4:
        try:
            # Allow swizzled getting of attrs
            vec_class = {2: Vec2, 3: Vec3, 4: Vec4}[len(attrs)]
            return vec_class(*(self['xy'.index(c)] for c in attrs))
        except (ValueError, KeyError, TypeError) as err:
            msg = f"'Vec2' has no attribute: '{attrs}'."
            raise AttributeError(msg) from err


class Vec3(_typing.NamedTuple):
    """A three-dimensional vector represented as X Y Z coordinates.

    `Vec3` is an immutable 3D Vector, including most common operators.
    As an immutable type, all operations return a new object.

    .. note:: The Python ``len`` operator returns the number of elements in
              the vector. For the vector length, use the `length()` method.
    """

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __bool__(self):
        return self.x != 0.0 or self.y != 0.0 or self.z != 0.0

    def __add__(self, other: Vec3 | tuple[float, float, float] | float) -> Vec3:
        try:
            return Vec3(
                self[0] + other[0], self[1] + other[1], self[2] + other[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] + other, self[1] + other, self[2] + other  # type: ignore
            )

    def __radd__(self, other: Vec3 | tuple[float, float, float] | float) -> Vec3:
        try:
            return self.__add__(_typing.cast(Vec3, other))
        except TypeError as err:
            if other == 0:  # Required for functionality with sum()
                return self
            raise err

    def __sub__(self, other: Vec3 | tuple[float, float, float] | float) -> Vec3:
        try:
            return Vec3(
                self[0] - other[0], self[1] - other[1], self[2] - other[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] - other, self[1] - other, self[2] - other  # type: ignore
            )

    def __rsub__(self, other: Vec3 | tuple[float, float, float] | float) -> Vec3:
        try:
            return Vec3(
                other[0] - self[0], other[1] - self[1], other[2] - self[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                other - self[0], other - self[1], other - self[2]  # type: ignore
            )

    def __mul__(self, scalar: float | Vec3 | tuple[float, float, float]) -> Vec3:
        try:
            return Vec3(
                self[0] * scalar[0], self[1] * scalar[1], self[2] * scalar[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] * scalar, self[1] * scalar, self[2] * scalar  # type: ignore
            )

    def __rmul__(self, scalar: float | Vec3 | tuple[float, float, float]) -> Vec3:
        try:
            return Vec3(
                self[0] * scalar[0], self[1] * scalar[1], self[2] * scalar[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] * scalar, self[1] * scalar, self[2] * scalar  # type: ignore
            )

    def __truediv__(self, scalar: float | Vec3 | tuple[float, float, float]) -> Vec3:
        try:
            return Vec3(
                self[0] / scalar[0], self[1] / scalar[1], self[2] / scalar[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] / scalar, self[1] / scalar, self[2] / scalar  # type: ignore
            )

    def __rtruediv__(self, other: float | Vec3 | tuple[float, float, float]) -> Vec3:
        try:
            return Vec3(
                other[0] / self[0], other[1] / self[1], other[2] / self[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                other / self[0], other / self[1], other / self[2]  # type: ignore
            )

    def __floordiv__(self, other: float | Vec3 | tuple[float, float, float]) -> Vec3:
        try:
            return Vec3(
                self[0] // other[0], self[1] // other[1], self[2] // other[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] // other, self[1] // other, self[2] // other  # type: ignore
            )

    def __rfloordiv__(self, other: float | Vec3 | tuple[float, float, float]) -> Vec3:
        try:
            return Vec3(
                other[0] // self[0], other[1] // self[1], other[2] // self[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                other // self[0], other // self[1], other // self[2]  # type: ignore
            )

    def __abs__(self) -> Vec3:
        return Vec3(abs(self[0]), abs(self[1]), abs(self[2]))

    def __neg__(self) -> Vec3:
        return Vec3(-self[0], -self[1], -self[2])

    def __round__(self, n_digits: int | None = None) -> Vec3:
        return Vec3(*(round(v, n_digits) for v in self))

    def __ceil__(self) -> Vec3:
        return Vec3(_math.ceil(self[0]), _math.ceil(self[1]), _math.ceil(self[2]))

    def __floor__(self) -> Vec3:
        return Vec3(_math.floor(self[0]), _math.floor(self[1]), _math.floor(self[2]))

    def __trunc__(self) -> Vec3:
        return Vec3(_math.trunc(self[0]), _math.trunc(self[1]), _math.trunc(self[2]))

    def __mod__(self, other: Vec3 | tuple[float, float, float] | float) -> Vec3:
        try:
            return Vec3(
                self[0] % other[0], self[1] % other[1], self[2] % other[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] % other, self[1] % other, self[2] % other  # type: ignore
            )

    def __pow__(self, other: Vec3 | tuple[float, float, float] | float) -> Vec3:
        try:
            return Vec3(
                self[0] ** other[0], self[1] ** other[1], self[2] ** other[2]  # type: ignore
            )
        except TypeError:
            return Vec3(
                self[0] ** other, self[1] ** other, self[2] ** other  # type: ignore
            )

    def __lt__(self, other: Vec3 | tuple[float, float, float]) -> bool:
        return self[0] ** 2 + self[1] ** 2 + self[2] ** 2 < other[0] ** 2 + other[1] ** 2 + other[2] ** 2

    @classmethod
    def from_pitch_yaw(cls, pitch: float, yaw: float) -> Vec3:
        """Create a unit vector from pitch and yaw in radians.

        Args:
            pitch: The pitch value in radians
            yaw: The yaw value in radians
        """
        return Vec3(
            _math.cos(yaw) * _math.cos(pitch),
            _math.sin(pitch),
            _math.sin(yaw) * _math.cos(pitch),
        ).normalize()

    def get_pitch_yaw(self) -> tuple[float, float]:
        """Get the pitch and yaw angles from a unit vector in radians."""
        return _math.asin(self.y), _math.atan2(self.z, self.x)

    def length(self) -> float:
        """Calculate the length of the vector: ``sqrt(x ** 2 + y ** 2 + z ** 2)``."""
        return _math.sqrt(self[0] ** 2 + self[1] ** 2 + self[2] ** 2)

    def length_squared(self) -> float:
        """Calculate the squared length of the vector.

        This is simply shortcut for `x ** 2 + y ** 2 + z ** 2` and can be used
        for faster comparisons without the need for a square root.
        """
        return self[0] ** 2 + self[1] ** 2 + self[2] ** 2

    def cross(self, other: Vec3 | tuple[float, float, float]) -> Vec3:
        """Calculate the cross product of this vector and another 3D vector.

        Args:
            other: Another Vec3 or tuple of 3 floats.
        """
        return Vec3(
            (self.y * other[2]) - (self.z * other[1]),
            (self.z * other[0]) - (self.x * other[2]),
            (self.x * other[1]) - (self.y * other[0]),
        )

    def dot(self, other: Vec3 | tuple[float, float, float]) -> float:
        """Calculate the dot product of this vector and another 3D vector.

        Args:
            other: Another Vec3 or tuple of 3 floats.
        """
        return self.x * other[0] + self.y * other[1] + self.z * other[2]

    def lerp(self, other: Vec3 | tuple[float, float, float], alpha: float) -> Vec3:
        """Create a new Vec3 linearly interpolated between this vector and another Vec3-like.

        Args:
          other: Another Vec3 instance or tuple of 3 floats.
          alpha: The amount of interpolation between this vector, and the other
                 vector. This should be a value between 0.0 and 1.0. For example:
                 0.5 is the midway point between both vectors.
        """
        return Vec3(
            self.x + (alpha * (other[0] - self.x)),
            self.y + (alpha * (other[1] - self.y)),
            self.z + (alpha * (other[2] - self.z)),
        )

    def distance(self, other: Vec3 | tuple[float, float, float]) -> float:
        """Calculate the distance between this vector and another 3D vector.

        Args:
            other: The point to calculate the distance to.
        """
        return _math.sqrt(((other[0] - self.x) ** 2) + ((other[1] - self.y) ** 2) + ((other[2] - self.z) ** 2))

    def normalize(self) -> Vec3:
        """Return a normalized version of the vector.

        This simply means the vector will have a length of 1.0. If the vector
        has a length of 0, the original vector will be returned.
        """
        try:
            d = _math.sqrt(self[0] ** 2 + self[1] ** 2 + self[2] ** 2)
            return Vec3(self.x / d, self.y / d, self.z / d)
        except ZeroDivisionError:
            return self

    if _typing.TYPE_CHECKING:
        @_typing.overload
        def clamp(self, min_val: float, max_val: float) -> Vec3:
            ...

        @_typing.overload
        def clamp(self, min_val: Vec3 | tuple[float, float, float], max_val: Vec3 | tuple[float, float, float]) -> Vec3:
            ...

        # -- Begin revert if perf-impacting block --
        @_typing.overload
        def clamp(self, min_val: Vec3 | tuple[float, float, float], max_val: float) -> Vec3:
            ...

        @_typing.overload
        def clamp(self, min_val: float, max_val: Vec3 | tuple[float, float, float]) -> Vec3:
            ...
        # -- End revert if perf-impacting block --

    def clamp(self, min_val: float | Vec3 | tuple[float, float, float],
              max_val: float | Vec3 | tuple[float, float, float]) -> Vec3:
        """Restrict the value of the X, Y and Z components of the vector to be within the given values.

        If a single value is provided, it will be used for all components.
        If a tuple or Vec3 is provided, the first value will be used for X, the second for Y, and the third for Z.

        Args:
            min_val: The minimum value(s)
            max_val: The maximum value(s)
        """
        try:
            min_x, min_y, min_z = min_val[0], min_val[1], min_val[2]  # type: ignore
        except TypeError:
            min_x = min_val
            min_y = min_val
            min_z = min_val
        try:
            max_x, max_y, max_z = max_val[0], max_val[1], max_val[2]  # type: ignore
        except TypeError:
            max_x = max_val
            max_y = max_val
            max_z = max_val

        return Vec3(
            clamp(self[0], min_x, max_x), clamp(self[1], min_y, max_y), clamp(self[2], min_z, max_z)  # type: ignore
        )

    def index(self, *args: _typing.Any) -> int:
        raise NotImplementedError("Vec types can be indexed directly.")

    def __getattr__(self, attrs: str) -> Vec2 | Vec3 | Vec4:
        try:
            # Allow swizzled getting of attrs
            vec_class = {2: Vec2, 3: Vec3, 4: Vec4}[len(attrs)]
            return vec_class(*(self['xyz'.index(c)] for c in attrs))
        except (ValueError, KeyError, TypeError) as err:
            msg = f"'Vec3' has no attribute: '{attrs}'."
            raise AttributeError(msg) from err


class Vec4(_typing.NamedTuple):
    """A four-dimensional vector represented as X Y Z W coordinates.

    `Vec4` is an immutable 4D Vector, including most common operators.
    As an immutable type, all operations return a new object.

    .. note:: The Python ``len` operator returns the number of elements in
              the vector. For the vector length, use the `length()` method.
    """

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 0.0

    def __bool__(self):
        return self.x != 0.0 or self.y != 0.0 or self.z != 0.0 or self.w != 0.0

    def __add__(self, other: Vec4 | tuple[float, float, float, float] | float) -> Vec4:
        try:
            return Vec4(
                other[0] + self[0], other[1] + self[1], other[2] + self[2], other[3] + self[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                other + self[0], other + self[1], other + self[2], other + self[3]  # type: ignore
            )

    def __radd__(self, other: Vec4 | tuple[float, float, float, float] | int) -> Vec4:
        try:
            return self.__add__(_typing.cast(Vec4, other))
        except TypeError as err:
            if other == 0:  # Required for functionality with sum()
                return self
            raise err

    def __sub__(self, other: Vec4 | tuple[float, float, float, float] | float) -> Vec4:
        try:
            return Vec4(
                self[0] - other[0], self[1] - other[1], self[2] - other[2], self[3] - other[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] - other, self[1] - other, self[2] - other, self[3] - other  # type: ignore
            )

    def __rsub__(self, other: Vec4 | tuple[float, float, float, float] | float) -> Vec4:
        try:
            return Vec4(
                other[0] - self[0], other[1] - self[1], other[2] - self[2], other[3] - self[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                other - self[0], other - self[1], other - self[2], other - self[3]  # type: ignore
            )

    def __mul__(self, scalar: float | Vec4 | tuple[float, float, float, float]) -> Vec4:
        try:
            return Vec4(
                self[0] * scalar[0], self[1] * scalar[1], self[2] * scalar[2], self[3] * scalar[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] * scalar, self[1] * scalar, self[2] * scalar, self[3] * scalar  # type: ignore
            )

    def __rmul__(self, scalar: float | Vec4 | tuple[float, float, float, float]) -> Vec4:
        try:
            return Vec4(
                self[0] * scalar[0], self[1] * scalar[1], self[2] * scalar[2], self[3] * scalar[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] * scalar, self[1] * scalar, self[2] * scalar, self[3] * scalar  # type: ignore
            )

    def __truediv__(self, scalar: float | Vec4 | tuple[float, float, float, float]) -> Vec4:
        try:
            return Vec4(
                self[0] / scalar[0], self[1] / scalar[1], self[2] / scalar[2], self[3] / scalar[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] / scalar, self[1] / scalar, self[2] / scalar, self[3] / scalar  # type: ignore
            )

    def __rtruediv__(self, scalar: float | Vec4 | tuple[float, float, float, float]) -> Vec4:
        try:
            return Vec4(
                scalar[0] / self[0], scalar[1] / self[1], scalar[2] / self[2], scalar[3] / self[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                scalar / self[0], scalar / self[1], scalar / self[2], scalar / self[3]  # type: ignore
            )

    def __floordiv__(self, scalar: float | Vec4 | tuple[float, float, float, float]) -> Vec4:
        try:
            return Vec4(
                self[0] // scalar[0], self[1] // scalar[1], self[2] // scalar[2], self[3] // scalar[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] // scalar, self[1] // scalar, self[2] // scalar, self[3] // scalar  # type: ignore
            )

    def __rfloordiv__(self, scalar: float | Vec4 | tuple[float, float, float, float]) -> Vec4:
        try:
            return Vec4(
                scalar[0] // self[0], scalar[1] // self[1], scalar[2] // self[2], scalar[3] // self[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                scalar // self[0], scalar // self[1], scalar // self[2], scalar // self[3]  # type: ignore
            )

    def __abs__(self) -> Vec4:
        return Vec4(abs(self[0]), abs(self[1]), abs(self[2]), abs(self[3]))

    def __neg__(self) -> Vec4:
        return Vec4(-self.x, -self.y, -self.z, -self.w)

    def __round__(self, n_digits: int | None = None) -> Vec4:
        return Vec4(*(round(v, n_digits) for v in self))

    def __ceil__(self) -> Vec4:
        return Vec4(_math.ceil(self[0]), _math.ceil(self[1]), _math.ceil(self[2]), _math.ceil(self[3]))

    def __floor__(self) -> Vec4:
        return Vec4(_math.floor(self[0]), _math.floor(self[1]), _math.floor(self[2]), _math.floor(self[3]))

    def __trunc__(self) -> Vec4:
        return Vec4(_math.trunc(self[0]), _math.trunc(self[1]), _math.trunc(self[2]), _math.trunc(self[3]))

    def __mod__(self, other: Vec4 | tuple[int, int, int, int] | float) -> Vec4:
        try:
            return Vec4(
                self[0] % other[0], self[1] % other[1], self[2] % other[2], self[3] % other[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] % other, self[1] % other, self[2] % other, self[3] % other  # type: ignore
            )

    def __pow__(self, other: Vec4 | tuple[int, int, int, int] | float) -> Vec4:
        try:
            return Vec4(
                self[0] ** other[0], self[1] ** other[1], self[2] ** other[2], self[3] ** other[3]  # type: ignore
            )
        except TypeError:
            return Vec4(
                self[0] ** other, self[1] ** other, self[2] ** other, self[3] ** other  # type: ignore
            )

    def __lt__(self, other: Vec4 | tuple[float, float, float, float]) -> bool:
        return (self[0] ** 2 + self[1] ** 2 + self[2] ** 2 + self[3] ** 2 <
                other[0] ** 2 + other[1] ** 2 + other[2] ** 2 + other[3] ** 2)

    def length(self) -> float:
        """Calculate the length of the vector: ``sqrt(x ** 2 + y ** 2 + z ** 2 + w ** 2)``."""
        return _math.sqrt(self[0] ** 2 + self[1] ** 2 + self[2] ** 2 + self[3] ** 2)

    def length_squared(self) -> float:
        """Calculate the squared length of the vector.

        This is simply shortcut for `x ** 2 + y ** 2 + z ** 2 + w ** 2` and can be used
        for faster comparisons without the need for a square root.
        """
        return self[0] ** 2 + self[1] ** 2 + self[2] ** 2 + self[3] ** 2

    def lerp(self, other: Vec4 | tuple[float, float, float, float], amount: float) -> Vec4:
        """Create a new Vec4 linearly interpolated between this vector and another Vec4.

        Args:
          other: Another Vec4 instance.
          amount: The amount of interpolation between this vector, and the other
                  vector. This should be a value between 0.0 and 1.0. For example:
                  0.5 is the midway point between both vectors.
        """
        return Vec4(
            self.x + (amount * (other.x - self.x)),
            self.y + (amount * (other.y - self.y)),
            self.z + (amount * (other.z - self.z)),
            self.w + (amount * (other.w - self.w)),
        )

    def distance(self, other: Vec4 | tuple[float, float, float, float]) -> float:
        """Calculate the distance between this vector and another 4D vector.

        Args:
            other: The point to calculate the distance to.
        """
        return _math.sqrt(
            ((other.x - self.x) ** 2)
            + ((other.y - self.y) ** 2)
            + ((other.z - self.z) ** 2)
            + ((other.w - self.w) ** 2),
        )

    def normalize(self) -> Vec4:
        """Returns a normalized version of the vector meaning a version that has length 1.

        This means that the vector will have the same direction, but a length of 1.0.
        If the vector has a length of 0, the original vector will be returned.
        """
        if d := _math.sqrt(self[0] ** 2 + self[1] ** 2 + self[2] ** 2 + self[3] ** 2):
            return Vec4(self[0] / d, self[1] / d, self[2] / d, self[3] / d)
        return self

    if _typing.TYPE_CHECKING:
        @_typing.overload
        def clamp(self, min_val: float, max_val: float) -> Vec4:
            ...

        @_typing.overload
        def clamp(self, min_val: tuple[float, float, float, float], max_val: tuple[float, float, float, float]) -> Vec4:
            ...

        # -- Begin revert if perf-impacting block --
        @_typing.overload
        def clamp(self, min_val: tuple[float, float, float, float], max_val: float) -> Vec4:
            ...

        @_typing.overload
        def clamp(self, min_val: float, max_val: tuple[float, float, float, float]) -> Vec4:
            ...
        # -- End revert if perf-impacting block --

    def clamp(
        self,
        min_val: float | tuple[float, float, float, float],
        max_val: float | tuple[float, float, float, float],
    ) -> Vec4:
        """Restrict the value of the X, Y, Z and W components of the vector to be within the given values.

        The minimum and maximum values can be a single value that will be applied to all components,
        or a tuple of 4 values that will be applied to each component respectively.

        Args:
            min_val: The minimum value(s)
            max_val: The maximum value(s)
        """
        try:
            min_x, min_y, min_z, min_w = min_val[0], min_val[1], min_val[2], min_val[3]  # type: ignore
        except TypeError:
            min_x = min_val
            min_y = min_val
            min_z = min_val
            min_w = min_val
        try:
            max_x, max_y, max_z, max_w = max_val[0], max_val[1],  max_val[2], max_val[3]  # type: ignore
        except TypeError:
            max_x = max_val
            max_y = max_val
            max_z = max_val
            max_w = max_val

        return Vec4(
            clamp(self[0], min_x, max_x), clamp(self[1], min_y, max_y),  # type: ignore
            clamp(self[2], min_z, max_z), clamp(self[3], min_w, max_w),  # type: ignore
        )

    def dot(self, other: Vec4 | tuple[float, float, float, float]) -> float:
        """Calculate the dot product of this vector and another 4D vector.

        Args:
            other: Another Vec4 instance.
        """
        return self.x * other.x + self.y * other.y + self.z * other.z + self.w * other.w

    def index(self, *args: _typing.Any) -> int:
        raise NotImplementedError("Vec types can be indexed directly.")

    def __getattr__(self, attrs: str) -> Vec2 | Vec3 | Vec4:
        try:
            # Allow swizzled getting of attrs
            vec_class = {2: Vec2, 3: Vec3, 4: Vec4}[len(attrs)]
            return vec_class(*(self['xyzw'.index(c)] for c in attrs))
        except (ValueError, KeyError, TypeError) as err:
            msg = f"'Vec4' has no attribute: '{attrs}'."
            raise AttributeError(msg) from err


class Mat3(_typing.NamedTuple):
    """A 3x3 Matrix.

    `Mat3` is an immutable 3x3 Matrix, which includes most common operators.

    A Matrix can be created with a list or tuple of 9 values.
    If no values are provided, an "identity matrix" will be created
    (1.0 on the main diagonal). Because Mat3 objects are immutable,
    all operations return a new Mat3 object.

    .. note:: Matrix multiplication is performed using the "@" operator.
    """

    a: float = 1.0
    b: float = 0.0
    c: float = 0.0

    d: float = 0.0
    e: float = 1.0
    f: float = 0.0

    g: float = 0.0
    h: float = 0.0
    i: float = 1.0

    def scale(self, sx: float, sy: float) -> Mat3:
        return self @ Mat3(1.0 / sx, 0.0, 0.0, 0.0, 1.0 / sy, 0.0, 0.0, 0.0, 1.0)

    def translate(self, tx: float, ty: float) -> Mat3:
        return self @ Mat3(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, -tx, ty, 1.0)

    def rotate(self, phi: float) -> Mat3:
        s = _math.sin(_math.radians(phi))
        c = _math.cos(_math.radians(phi))
        return self @ Mat3(c, s, 0.0, -s, c, 0.0, 0.0, 0.0, 1.0)

    def shear(self, sx: float, sy: float) -> Mat3:
        return self @ Mat3(1.0, sy, 0.0, sx, 1.0, 0.0, 0.0, 0.0, 1.0)

    def __add__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError("Can only add to other Mat3 types")
        return Mat3(*(s + o for s, o in zip(self, other)))

    def __sub__(self, other: Mat3) -> Mat3:
        if not isinstance(other, Mat3):
            raise TypeError("Can only subtract from other Mat3 types")
        return Mat3(*(s - o for s, o in zip(self, other)))

    def __pos__(self) -> Mat3:
        return self

    def __neg__(self) -> Mat3:
        return Mat3(*(-v for v in self))

    def __invert__(self) -> Mat3:
        # extract the elements in row-column form. (matrix is stored column first)
        a11, a12, a13, a21, a22, a23, a31, a32, a33 = self

        # Calculate Adj(self) values column-row order
        # | a d g |
        # | b e h |
        # | c f i |
        a = a22 * a33 - a32 * a23  # +
        b = a31 * a23 - a21 * a33  # -
        c = a21 * a32 - a22 * a31  # +
        d = a32 * a13 - a12 * a33  # -
        e = a11 * a33 - a31 * a13  # +
        f = a31 * a12 - a11 * a32  # -
        g = a12 * a23 - a22 * a13  # +
        h = a21 * a13 - a11 * a23  # -
        i = a11 * a22 - a21 * a12  # +

        # Calculate determinant
        det = a11 * a + a21 * d + a31 * g

        if det == 0:
            _warnings.warn("Unable to calculate inverse of singular Matrix")
            return self

        # get determinant reciprocal
        rep = 1.0 / det

        # get inverse: A^-1 = def(A)^-1 * adj(A)
        return Mat3(*(a * rep, b * rep, c * rep,
                      d * rep, e * rep, f * rep,
                      g * rep, h * rep, i * rep))

    def __round__(self, ndigits: int | None = None) -> Mat3:
        return Mat3(*(round(v, ndigits) for v in self))

    def __mul__(self, other: object) -> _typing.NoReturn:
        msg = "Please use the @ operator for Matrix multiplication."
        raise NotImplementedError(msg)

    @_typing.overload
    def __matmul__(self, other: Vec3) -> Vec3: ...

    @_typing.overload
    def __matmul__(self, other: Mat3) -> Mat3: ...

    def __matmul__(self, other) -> Vec3 | Mat3:
        try:
            # extract the elements in row-column form. (matrix is stored column first)
            a11, a12, a13, a21, a22, a23, a31, a32, a33 = self
            b11, b12, b13, b21, b22, b23, b31, b32, b33 = other

            # Multiply and sum rows * columns
            return Mat3(
                # Column 1
                a11 * b11 + a21 * b12 + a31 * b13, a12 * b11 + a22 * b12 + a32 * b13, a13 * b11 + a23 * b12 + a33 * b13,
                # Column 2
                a11 * b21 + a21 * b22 + a31 * b23, a12 * b21 + a22 * b22 + a32 * b23, a13 * b21 + a23 * b22 + a33 * b23,
                # Column 3
                a11 * b31 + a21 * b32 + a31 * b33, a12 * b31 + a22 * b32 + a32 * b33, a13 * b31 + a23 * b32 + a33 * b33,
            )
        except ValueError:
            x, y, z = other
            # extract the elements in row-column form. (matrix is stored column first)
            a11, a12, a13, a21, a22, a23, a31, a32, a33 = self
            return Vec3(
                a11 * x + a21 * y + a31 * z,
                a12 * x + a22 * y + a32 * z,
                a13 * x + a23 * y + a33 * z,
            )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self[0:3]}\n    {self[3:6]}\n    {self[6:9]}"


class Mat4(_typing.NamedTuple):
    """A 4x4 Matrix.

    ``Mat4`` is an immutable 4x4 Matrix object with most common operators.
    This includes class methods for creating orthogonal and perspective
    projection matrixes, which can be used directly by OpenGL.

    You can create a Matrix with 16 initial values (floats), or no values
    at all. If no values are provided, an "identity matrix" will be created
    (1.0 on the main diagonal). ``Mat4`` objects are immutable, so all
    operations will return a new Mat4 object.

    Note:
        Matrix multiplication is performed using the "@" operator.
    """

    a: float = 1.0
    b: float = 0.0
    c: float = 0.0
    d: float = 0.0

    e: float = 0.0
    f: float = 1.0
    g: float = 0.0
    h: float = 0.0

    i: float = 0.0
    j: float = 0.0
    k: float = 1.0
    l: float = 0.0

    m: float = 0.0
    n: float = 0.0
    o: float = 0.0
    p: float = 1.0

    @classmethod
    def orthogonal_projection(cls: type[Mat4], left: float, right: float, bottom: float, top: float,
                              z_near: float, z_far: float) -> Mat4:
        """Create a Mat4 orthographic projection matrix for use with OpenGL.

        Given left, right, bottom, top values, and near/far z planes,
        create a 4x4 Projection Matrix. This is useful for setting
        :py:attr:`~pyglet.window.Window.projection`.
        """
        width = right - left
        height = top - bottom
        depth = z_far - z_near

        s_x = 2.0 / width
        s_y = 2.0 / height
        s_z = 2.0 / -depth

        t_x = -(right + left) / width
        t_y = -(top + bottom) / height
        t_z = -(z_far + z_near) / depth

        return cls(s_x, 0.0, 0.0, 0.0,
                   0.0, s_y, 0.0, 0.0,
                   0.0, 0.0, s_z, 0.0,
                   t_x, t_y, t_z, 1.0)

    @classmethod
    def perspective_projection(cls: type[Mat4], aspect: float, z_near: float, z_far: float, fov: float = 60) -> Mat4:
        """Create a Mat4 perspective projection matrix for use with OpenGL.

        Given a desired aspect ratio, near/far planes, and fov (field of view),
        create a 4x4 Projection Matrix. This is useful for setting
        :py:attr:`~pyglet.window.Window.projection`.
        """
        xy_max = z_near * _math.tan(fov * _math.pi / 360)
        y_min = -xy_max
        x_min = -xy_max

        width = xy_max - x_min
        height = xy_max - y_min
        depth = z_far - z_near
        q = -(z_far + z_near) / depth
        qn = -2 * z_far * z_near / depth

        w = 2 * z_near / width
        w = w / aspect
        h = 2 * z_near / height

        return cls(w, 0, 0, 0,
                   0, h, 0, 0,
                   0, 0, q, -1,
                   0, 0, qn, 0)

    @classmethod
    def from_rotation(cls, angle: float, vector: Vec3) -> Mat4:
        """Create a rotation matrix from an angle and Vec3.

        Args:
          angle: The desired angle, in radians.
          vector: A Vec3 indicating the direction.
        """
        return cls().rotate(angle, vector)

    @classmethod
    def from_scale(cls: type[Mat4], vector: Vec3) -> Mat4:
        """Create a scale matrix from a Vec3."""
        return cls(vector.x, 0.0, 0.0, 0.0,
                   0.0, vector.y, 0.0, 0.0,
                   0.0, 0.0, vector.z, 0.0,
                   0.0, 0.0, 0.0, 1.0)

    @classmethod
    def from_translation(cls: type[Mat4], vector: Vec3) -> Mat4:
        """Create a translation matrix from a Vec3."""
        return cls(1.0, 0.0, 0.0, 0.0,
                   0.0, 1.0, 0.0, 0.0,
                   0.0, 0.0, 1.0, 0.0,
                   vector.x, vector.y, vector.z, 1.0)

    @classmethod
    def look_at(cls: type[Mat4], position: Vec3, target: Vec3, up: Vec3) -> Mat4:
        """Create a viewing matrix that points toward a target.

        This method takes three Vec3s, describing the viewer's position,
        where they are looking, and the upward axis (typically positive
        on the Y axis). The resulting Mat4 can be used as the projection
        matrix.

        Args:
          position: The location of the viewer in the scene.
          target: The point that the viewer is looking towards.
          up: A vector pointing "up" in the scene, typically `Vec3(0.0, 1.0, 0.0)`.
        """
        f = (target - position).normalize()
        u = up.normalize()
        s = f.cross(u).normalize()
        u = s.cross(f)

        return cls(s.x, u.x, -f.x, 0.0,
                   s.y, u.y, -f.y, 0.0,
                   s.z, u.z, -f.z, 0.0,
                   -s.dot(position), -u.dot(position), f.dot(position), 1.0)

    def row(self, index: int) -> tuple:
        """Get a specific row as a tuple."""
        return self[index::4]

    def column(self, index: int) -> tuple:
        """Get a specific column as a tuple."""
        return self[index * 4: index * 4 + 4]

    def rotate(self, angle: float, vector: Vec3) -> Mat4:
        """Get a rotation Matrix on x, y, or z axis."""
        if not all(abs(n) <= 1 for n in vector):
            raise ValueError("vector must be normalized (<=1)")
        x, y, z = vector
        c = _math.cos(angle)
        s = _math.sin(angle)
        t = 1 - c
        temp_x, temp_y, temp_z = t * x, t * y, t * z

        ra = c + temp_x * x
        rb = 0 + temp_x * y + s * z
        rc = 0 + temp_x * z - s * y
        re = 0 + temp_y * x - s * z
        rf = c + temp_y * y
        rg = 0 + temp_y * z + s * x
        ri = 0 + temp_z * x + s * y
        rj = 0 + temp_z * y - s * x
        rk = c + temp_z * z

        # ra, rb, rc, --
        # re, rf, rg, --
        # ri, rj, rk, --
        # --, --, --, --

        return self @ Mat4(ra, rb, rc, 0, re, rf, rg, 0, ri, rj, rk, 0, 0, 0, 0, 1)

    def scale(self, vector: Vec3) -> Mat4:
        """Get a scale Matrix on x, y, or z axis."""
        temp = list(self)
        temp[0] *= vector[0]
        temp[5] *= vector[1]
        temp[10] *= vector[2]
        return Mat4(*temp)

    def translate(self, vector: Vec3) -> Mat4:
        """Get a translation Matrix along x, y, and z axis."""
        return self @ Mat4(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, *vector, 1)

    def transpose(self) -> Mat4:
        """Get a transpose of this Matrix."""
        return Mat4(*self[0::4], *self[1::4], *self[2::4], *self[3::4])

    def __add__(self, other: Mat4) -> Mat4:
        if not isinstance(other, Mat4):
            raise TypeError("Can only add to other Mat4 types")
        return Mat4(*(s + o for s, o in zip(self, other)))

    def __sub__(self, other: Mat4) -> Mat4:
        if not isinstance(other, Mat4):
            raise TypeError("Can only subtract from other Mat4 types")
        return Mat4(*(s - o for s, o in zip(self, other)))

    def __pos__(self) -> Mat4:
        return self

    def __neg__(self) -> Mat4:
        return Mat4(*(-v for v in self))

    def __invert__(self) -> Mat4:
        # extract the elements in row-column form. (matrix is stored column first)
        a11, a12, a13, a14, a21, a22, a23, a24, a31, a32, a33, a34, a41, a42, a43, a44 = self

        a = a33 * a44 - a34 * a43
        b = a32 * a44 - a34 * a42
        c = a32 * a43 - a33 * a42
        d = a31 * a44 - a34 * a41
        e = a31 * a43 - a33 * a41
        f = a31 * a42 - a32 * a41
        g = a23 * a44 - a24 * a43
        h = a22 * a44 - a24 * a42
        i = a22 * a43 - a23 * a42
        j = a23 * a34 - a24 * a33
        k = a22 * a34 - a24 * a32
        l = a22 * a33 - a23 * a32
        m = a21 * a44 - a24 * a41
        n = a21 * a43 - a23 * a41
        o = a21 * a34 - a24 * a31
        p = a21 * a33 - a23 * a31
        q = a21 * a42 - a22 * a41
        r = a21 * a32 - a22 * a31

        det = (a11 * (a22 * a - a23 * b + a24 * c) -
               a12 * (a21 * a - a23 * d + a24 * e) +
               a13 * (a21 * b - a22 * d + a24 * f) -
               a14 * (a21 * c - a22 * e + a23 * f))

        if det == 0:
            _warnings.warn("Unable to calculate inverse of singular Matrix")
            return self

        pdet = 1 / det
        ndet = -pdet

        return Mat4(pdet * (a22 * a - a23 * b + a24 * c),
                    ndet * (a12 * a - a13 * b + a14 * c),
                    pdet * (a12 * g - a13 * h + a14 * i),
                    ndet * (a12 * j - a13 * k + a14 * l),
                    ndet * (a21 * a - a23 * d + a24 * e),
                    pdet * (a11 * a - a13 * d + a14 * e),
                    ndet * (a11 * g - a13 * m + a14 * n),
                    pdet * (a11 * j - a13 * o + a14 * p),
                    pdet * (a21 * b - a22 * d + a24 * f),
                    ndet * (a11 * b - a12 * d + a14 * f),
                    pdet * (a11 * h - a12 * m + a14 * q),
                    ndet * (a11 * k - a12 * o + a14 * r),
                    ndet * (a21 * c - a22 * e + a23 * f),
                    pdet * (a11 * c - a12 * e + a13 * f),
                    ndet * (a11 * i - a12 * n + a13 * q),
                    pdet * (a11 * l - a12 * p + a13 * r))

    def __round__(self, ndigits: int | None) -> Mat4:
        return Mat4(*(round(v, ndigits) for v in self))

    def __mul__(self, other: int) -> _typing.NoReturn:
        msg = "Please use the @ operator for Matrix multiplication."
        raise NotImplementedError(msg)

    @_typing.overload
    def __matmul__(self, other: Vec4) -> Vec4: ...

    @_typing.overload
    def __matmul__(self, other: Mat4) -> Mat4: ...

    def __matmul__(self, other):
        try:
            # extract the elements in row-column form. (matrix is stored column first)
            a11, a12, a13, a14, a21, a22, a23, a24, a31, a32, a33, a34, a41, a42, a43, a44 = self
            b11, b12, b13, b14, b21, b22, b23, b24, b31, b32, b33, b34, b41, b42, b43, b44 = other
            # Multiply and sum rows * columns:
            return Mat4(
                # Column 1
                a11 * b11 + a21 * b12 + a31 * b13 + a41 * b14, a12 * b11 + a22 * b12 + a32 * b13 + a42 * b14,
                a13 * b11 + a23 * b12 + a33 * b13 + a43 * b14, a14 * b11 + a24 * b12 + a34 * b13 + a44 * b14,
                # Column 2
                a11 * b21 + a21 * b22 + a31 * b23 + a41 * b24, a12 * b21 + a22 * b22 + a32 * b23 + a42 * b24,
                a13 * b21 + a23 * b22 + a33 * b23 + a43 * b24, a14 * b21 + a24 * b22 + a34 * b23 + a44 * b24,
                # Column 3
                a11 * b31 + a21 * b32 + a31 * b33 + a41 * b34, a12 * b31 + a22 * b32 + a32 * b33 + a42 * b34,
                a13 * b31 + a23 * b32 + a33 * b33 + a43 * b34, a14 * b31 + a24 * b32 + a34 * b33 + a44 * b34,
                # Column 4
                a11 * b41 + a21 * b42 + a31 * b43 + a41 * b44, a12 * b41 + a22 * b42 + a32 * b43 + a42 * b44,
                a13 * b41 + a23 * b42 + a33 * b43 + a43 * b44, a14 * b41 + a24 * b42 + a34 * b43 + a44 * b44,
            )
        except ValueError:
            x, y, z, w = other
            # extract the elements in row-column form. (matrix is stored column first)
            a11, a12, a13, a14, a21, a22, a23, a24, a31, a32, a33, a34, a41, a42, a43, a44 = self
            return Vec4(
                x * a11 + y * a21 + z * a31 + w * a41,
                x * a12 + y * a22 + z * a32 + w * a42,
                x * a13 + y * a23 + z * a33 + w * a43,
                x * a14 + y * a24 + z * a34 + w * a44,
            )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self[0:4]}\n    {self[4:8]}\n    {self[8:12]}\n    {self[12:16]}"


class Quaternion(_typing.NamedTuple):
    """Quaternion."""

    w: float = 1.0
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    @classmethod
    def from_mat3(cls) -> Quaternion:
        raise NotImplementedError

    @classmethod
    def from_mat4(cls) -> Quaternion:
        raise NotImplementedError

    def to_mat4(self) -> Mat4:
        w = self.w
        x = self.x
        y = self.y
        z = self.z

        a = 1 - (y**2 + z**2) * 2
        b = 2 * (x * y - z * w)
        c = 2 * (x * z + y * w)

        e = 2 * (x * y + z * w)
        f = 1 - (x**2 + z**2) * 2
        g = 2 * (y * z - x * w)

        i = 2 * (x * z - y * w)
        j = 2 * (y * z + x * w)
        k = 1 - (x**2 + y**2) * 2

        # a, b, c, -
        # e, f, g, -
        # i, j, k, -
        # -, -, -, -

        return Mat4(a, b, c, 0.0, e, f, g, 0.0, i, j, k, 0.0, 0.0, 0.0, 0.0, 1.0)

    def to_mat3(self) -> Mat3:
        w = self.w
        x = self.x
        y = self.y
        z = self.z

        a = 1 - (y**2 + z**2) * 2
        b = 2 * (x * y - z * w)
        c = 2 * (x * z + y * w)

        e = 2 * (x * y + z * w)
        f = 1 - (x**2 + z**2) * 2
        g = 2 * (y * z - x * w)

        i = 2 * (x * z - y * w)
        j = 2 * (y * z + x * w)
        k = 1 - (x**2 + y**2) * 2

        # a, b, c, -
        # e, f, g, -
        # i, j, k, -
        # -, -, -, -

        return Mat3(*(a, b, c, e, f, g, i, j, k))

    def length(self) -> float:
        """Calculate the length of the Quaternion.

        The distance between the coordinates and the origin.
        """
        return _math.sqrt(self.w**2 + self.x**2 + self.y**2 + self.z**2)

    def conjugate(self) -> Quaternion:
        return Quaternion(self.w, -self.x, -self.y, -self.z)

    def dot(self, other: Quaternion) -> float:
        a, b, c, d = self
        e, f, g, h = other
        return a * e + b * f + c * g + d * h

    def normalize(self) -> Quaternion:
        m = self.length()
        if m == 0:
            return self
        return Quaternion(self.w / m, self.x / m, self.y / m, self.z / m)

    def __add__(self, other: Quaternion) -> Quaternion:
        a, b, c, d = self
        e, f, g, h = other
        return Quaternion(a + e, b + f, c + g, d + h)

    def __sub__(self, other: Quaternion) -> Quaternion:
        a, b, c, d = self
        e, f, g, h = other
        return Quaternion(a - e, b - f, c - g, d - h)

    def __mul__(self, scalar: float) -> Quaternion:
        w, x = self.w * scalar, self.x * scalar
        y, z = self.y * scalar, self.z * scalar
        return Quaternion(w, x, y, z)

    def __truediv__(self, other: Quaternion) -> Quaternion:
        return ~self @ other

    def __invert__(self) -> Quaternion:
        return self.conjugate() * (1 / self.dot(self))

    def __matmul__(self, other: Quaternion) -> Quaternion:
        a, u = self.w, Vec3(*self[1:])
        b, v = other.w, Vec3(*other[1:])
        scalar = a * b - u.dot(v)
        vector = v * a + u * b + u.cross(v)
        return Quaternion(scalar, *vector)



#!/usr/bin/env python
"""A full-screen minute:second timer.  Leave it in charge of your conference
lighting talks.

After 5 minutes, the timer goes red.  This limit is easily adjustable by
hacking the source code.

Press spacebar to start, stop and reset the timer.
"""

import pyglet

window = pyglet.window.Window(fullscreen=True)


class Timer:
    def __init__(self):
        self.label = pyglet.text.Label('00:00', font_size=360,
                                       x=window.width//2, y=window.height//2,
                                       anchor_x='center', anchor_y='center')
        self.reset()

    def reset(self):
        self.time = 0
        self.running = False
        self.label.text = '00:00'
        self.label.color = (255, 255, 255, 255)

    def update(self, dt):
        if self.running:
            self.time += dt
            m, s = divmod(self.time, 60)
            self.label.text = f'{round(m):02}:{round(s):02}'
            if m >= 5:
                self.label.color = (180, 0, 0, 255)

        window.draw(dt=dt)


@window.event
def on_key_press(symbol, modifiers):
    if symbol == pyglet.window.key.SPACE:
        if timer.running:
            timer.running = False
        else:
            if timer.time > 0:
                timer.reset()
            else:
                timer.running = True
    elif symbol == pyglet.window.key.ESCAPE:
        window.close()


@window.event
def on_draw():
    window.clear()
    timer.label.draw()


timer = Timer()

# Set timer and redraw update rate to 30 FPS
pyglet.clock.schedule_interval(timer.update, 1/30.0)

# Launch the application
pyglet.app.run(None)




import ctypes
import sys
from ctypes import (
    POINTER,
    WINFUNCTYPE,
    Structure,
    Union,
    c_byte,
    c_char,
    c_char_p,
    c_int,
    c_int8,
    c_int16,
    c_int32,
    c_int64,
    c_long,
    c_short,
    c_size_t,
    c_ubyte,
    c_uint,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
    c_wchar,
    c_wchar_p,
    cast,
    sizeof,
)
from ctypes.wintypes import (
    BOOL,
    BYTE,
    DOUBLE,  # noqa: F401
    DWORD,
    FILETIME,
    HANDLE,
    HBITMAP,
    HBRUSH,
    HDC,
    HICON,
    HINSTANCE,
    HMONITOR,
    HRGN,
    HWND,
    LARGE_INTEGER,
    LONG,
    LPARAM,
    LPCWSTR,
    LPOLESTR,
    LPWSTR,
    MSG,
    POINT,
    POINTL,
    RECT,
    UINT,
    ULARGE_INTEGER,
    ULONG,
    USHORT,
    WCHAR,
    WORD,
    WPARAM,
)

from . import com

_int_types = (c_int16, c_int32)
if hasattr(ctypes, 'c_int64'):
    # Some builds of ctypes apparently do not have c_int64
    # defined; it's a pretty good bet that these builds do not
    # have 64-bit pointers.
    _int_types += (c_int64,)
for t in _int_types:
    if sizeof(t) == sizeof(c_size_t):
        c_ptrdiff_t = t
del t
del _int_types


class c_void(Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', c_int)]


def POINTER_(obj):
    p = ctypes.POINTER(obj)

    # Convert None to a real NULL pointer to work around bugs
    # in how ctypes handles None on 64-bit platforms
    if not isinstance(p.from_param, classmethod):
        def from_param(cls, x):
            if x is None:
                return cls()
            else:
                return x

        p.from_param = classmethod(from_param)

    return p


c_void_p = POINTER_(c_void)

HCURSOR = HANDLE
LRESULT = LPARAM
COLORREF = DWORD
PVOID = c_void_p
BCHAR = c_wchar
LPRECT = POINTER(RECT)
LPPOINT = POINTER(POINT)
LPMSG = POINTER(MSG)
UINT_PTR = HANDLE
LONG_PTR = HANDLE
HDROP = HANDLE
LPTSTR = LPWSTR
LPSTREAM = c_void_p
CLSID = com.GUID

INT8 = c_int8
UINT8 = c_uint8
INT16 = c_int16
UINT16 = c_uint16
INT32 = c_int32
UINT32 = c_uint32
INT64 = c_int64
UINT64 = c_uint64
VOID = None

# Fixed in python 3.12. Ctypes defined BYTE as c_byte in previous versions.
# Ensure it's the same across all versions.
if sys.version_info < (3, 12):
    BYTE = c_ubyte

LF_FACESIZE = 32
CCHDEVICENAME = 32
CCHFORMNAME = 32

WNDPROC = WINFUNCTYPE(LRESULT, HWND, UINT, WPARAM, LPARAM)
TIMERPROC = WINFUNCTYPE(None, HWND, UINT, POINTER(UINT), DWORD)
TIMERAPCPROC = WINFUNCTYPE(None, PVOID, DWORD, DWORD)
MONITORENUMPROC = WINFUNCTYPE(BOOL, HMONITOR, HDC, LPRECT, LPARAM)

PROCESS_DPI_AWARENESS = UINT
PROCESS_DPI_UNAWARE = 0
PROCESS_SYSTEM_DPI_AWARE = 1
PROCESS_PER_MONITOR_DPI_AWARE = 2

MONITOR_DPI_TYPE = UINT
MDT_EFFECTIVE_DPI = 0
MDT_ANGULAR_DPI = 1
MDT_RAW_DPI = 2
MDT_DEFAULT = 3

DPI_AWARENESS_CONTEXT = HANDLE
DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2 = DPI_AWARENESS_CONTEXT(-4)

MONITOR_DEFAULTTONEAREST = 0
MONITOR_DEFAULTTONULL = 1
MONITOR_DEFAULTTOPRIMARY = 2

def MAKEINTRESOURCE(i):
    return cast(ctypes.c_void_p(i & 0xFFFF), c_wchar_p)


class WNDCLASS(Structure):
    _fields_ = [
        ('style', UINT),
        ('lpfnWndProc', WNDPROC),
        ('cbClsExtra', c_int),
        ('cbWndExtra', c_int),
        ('hInstance', HINSTANCE),
        ('hIcon', HICON),
        ('hCursor', HCURSOR),
        ('hbrBackground', HBRUSH),
        ('lpszMenuName', c_char_p),
        ('lpszClassName', c_wchar_p)
    ]


class SECURITY_ATTRIBUTES(Structure):
    _fields_ = [
        ("nLength", DWORD),
        ("lpSecurityDescriptor", c_void_p),
        ("bInheritHandle", BOOL)
    ]
    __slots__ = [f[0] for f in _fields_]


class PIXELFORMATDESCRIPTOR(Structure):
    _fields_ = [
        ('nSize', WORD),
        ('nVersion', WORD),
        ('dwFlags', DWORD),
        ('iPixelType', BYTE),
        ('cColorBits', BYTE),
        ('cRedBits', BYTE),
        ('cRedShift', BYTE),
        ('cGreenBits', BYTE),
        ('cGreenShift', BYTE),
        ('cBlueBits', BYTE),
        ('cBlueShift', BYTE),
        ('cAlphaBits', BYTE),
        ('cAlphaShift', BYTE),
        ('cAccumBits', BYTE),
        ('cAccumRedBits', BYTE),
        ('cAccumGreenBits', BYTE),
        ('cAccumBlueBits', BYTE),
        ('cAccumAlphaBits', BYTE),
        ('cDepthBits', BYTE),
        ('cStencilBits', BYTE),
        ('cAuxBuffers', BYTE),
        ('iLayerType', BYTE),
        ('bReserved', BYTE),
        ('dwLayerMask', DWORD),
        ('dwVisibleMask', DWORD),
        ('dwDamageMask', DWORD)
    ]


class RGBQUAD(Structure):
    _fields_ = [
        ('rgbBlue', BYTE),
        ('rgbGreen', BYTE),
        ('rgbRed', BYTE),
        ('rgbReserved', BYTE),
    ]
    __slots__ = [f[0] for f in _fields_]


class CIEXYZ(Structure):
    _fields_ = [
        ('ciexyzX', DWORD),
        ('ciexyzY', DWORD),
        ('ciexyzZ', DWORD),
    ]
    __slots__ = [f[0] for f in _fields_]


class CIEXYZTRIPLE(Structure):
    _fields_ = [
        ('ciexyzRed', CIEXYZ),
        ('ciexyzBlue', CIEXYZ),
        ('ciexyzGreen', CIEXYZ),
    ]
    __slots__ = [f[0] for f in _fields_]


class BITMAPINFOHEADER(Structure):
    _fields_ = [
        ('biSize', DWORD),
        ('biWidth', LONG),
        ('biHeight', LONG),
        ('biPlanes', WORD),
        ('biBitCount', WORD),
        ('biCompression', DWORD),
        ('biSizeImage', DWORD),
        ('biXPelsPerMeter', LONG),
        ('biYPelsPerMeter', LONG),
        ('biClrUsed', DWORD),
        ('biClrImportant', DWORD),
    ]


class BITMAPV5HEADER(Structure):
    _fields_ = [
        ('bV5Size', DWORD),
        ('bV5Width', LONG),
        ('bV5Height', LONG),
        ('bV5Planes', WORD),
        ('bV5BitCount', WORD),
        ('bV5Compression', DWORD),
        ('bV5SizeImage', DWORD),
        ('bV5XPelsPerMeter', LONG),
        ('bV5YPelsPerMeter', LONG),
        ('bV5ClrUsed', DWORD),
        ('bV5ClrImportant', DWORD),
        ('bV5RedMask', DWORD),
        ('bV5GreenMask', DWORD),
        ('bV5BlueMask', DWORD),
        ('bV5AlphaMask', DWORD),
        ('bV5CSType', DWORD),
        ('bV5Endpoints', CIEXYZTRIPLE),
        ('bV5GammaRed', DWORD),
        ('bV5GammaGreen', DWORD),
        ('bV5GammaBlue', DWORD),
        ('bV5Intent', DWORD),
        ('bV5ProfileData', DWORD),
        ('bV5ProfileSize', DWORD),
        ('bV5Reserved', DWORD),
    ]


class BITMAPINFO(Structure):
    _fields_ = [
        ('bmiHeader', BITMAPINFOHEADER),
        ('bmiColors', RGBQUAD * 1)
    ]
    __slots__ = [f[0] for f in _fields_]


class LOGFONT(Structure):
    _fields_ = [
        ('lfHeight', LONG),
        ('lfWidth', LONG),
        ('lfEscapement', LONG),
        ('lfOrientation', LONG),
        ('lfWeight', LONG),
        ('lfItalic', BYTE),
        ('lfUnderline', BYTE),
        ('lfStrikeOut', BYTE),
        ('lfCharSet', BYTE),
        ('lfOutPrecision', BYTE),
        ('lfClipPrecision', BYTE),
        ('lfQuality', BYTE),
        ('lfPitchAndFamily', BYTE),
        ('lfFaceName', (c_char * LF_FACESIZE))  # Use ASCII
    ]


class LOGFONTW(Structure):
    # https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-logfontw
    _fields_ = [
        ('lfHeight', LONG),
        ('lfWidth', LONG),
        ('lfEscapement', LONG),
        ('lfOrientation', LONG),
        ('lfWeight', LONG),
        ('lfItalic', BYTE),
        ('lfUnderline', BYTE),
        ('lfStrikeOut', BYTE),
        ('lfCharSet', BYTE),
        ('lfOutPrecision', BYTE),
        ('lfClipPrecision', BYTE),
        ('lfQuality', BYTE),
        ('lfPitchAndFamily', BYTE),
        ('lfFaceName', (WCHAR * LF_FACESIZE))
    ]


class TRACKMOUSEEVENT(Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('dwFlags', DWORD),
        ('hwndTrack', HWND),
        ('dwHoverTime', DWORD)
    ]
    __slots__ = [f[0] for f in _fields_]


class MINMAXINFO(Structure):
    _fields_ = [
        ('ptReserved', POINT),
        ('ptMaxSize', POINT),
        ('ptMaxPosition', POINT),
        ('ptMinTrackSize', POINT),
        ('ptMaxTrackSize', POINT)
    ]
    __slots__ = [f[0] for f in _fields_]


class ABC(Structure):
    _fields_ = [
        ('abcA', c_int),
        ('abcB', c_uint),
        ('abcC', c_int)
    ]
    __slots__ = [f[0] for f in _fields_]


class TEXTMETRIC(Structure):
    _fields_ = [
        ('tmHeight', c_long),
        ('tmAscent', c_long),
        ('tmDescent', c_long),
        ('tmInternalLeading', c_long),
        ('tmExternalLeading', c_long),
        ('tmAveCharWidth', c_long),
        ('tmMaxCharWidth', c_long),
        ('tmWeight', c_long),
        ('tmOverhang', c_long),
        ('tmDigitizedAspectX', c_long),
        ('tmDigitizedAspectY', c_long),
        ('tmFirstChar', c_char),  # Use ASCII
        ('tmLastChar', c_char),
        ('tmDefaultChar', c_char),
        ('tmBreakChar', c_char),
        ('tmItalic', c_byte),
        ('tmUnderlined', c_byte),
        ('tmStruckOut', c_byte),
        ('tmPitchAndFamily', c_byte),
        ('tmCharSet', c_byte)
    ]
    __slots__ = [f[0] for f in _fields_]


class MONITORINFOEX(Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('rcMonitor', RECT),
        ('rcWork', RECT),
        ('dwFlags', DWORD),
        ('szDevice', WCHAR * CCHDEVICENAME)
    ]
    __slots__ = [f[0] for f in _fields_]


class _DUMMYSTRUCTNAME(Structure):
    _fields_ = [
        ('dmOrientation', c_short),
        ('dmPaperSize', c_short),
        ('dmPaperLength', c_short),
        ('dmPaperWidth', c_short),
        ('dmScale', c_short),
        ('dmCopies', c_short),
        ('dmDefaultSource', c_short),
        ('dmPrintQuality', c_short),
    ]


class _DUMMYSTRUCTNAME2(Structure):
    _fields_ = [
        ('dmPosition', POINTL),
        ('dmDisplayOrientation', DWORD),
        ('dmDisplayFixedOutput', DWORD)
    ]


class _DUMMYDEVUNION(Union):
    _anonymous_ = ('_dummystruct1', '_dummystruct2')
    _fields_ = [
        ('_dummystruct1', _DUMMYSTRUCTNAME),
        ('dmPosition', POINTL),
        ('_dummystruct2', _DUMMYSTRUCTNAME2),
    ]


class DEVMODE(Structure):
    _anonymous_ = ('_dummyUnion',)
    _fields_ = [
        ('dmDeviceName', BCHAR * CCHDEVICENAME),
        ('dmSpecVersion', WORD),
        ('dmDriverVersion', WORD),
        ('dmSize', WORD),
        ('dmDriverExtra', WORD),
        ('dmFields', DWORD),
        # Just using the largest union member here
        ('_dummyUnion', _DUMMYDEVUNION),
        # End union
        ('dmColor', c_short),
        ('dmDuplex', c_short),
        ('dmYResolution', c_short),
        ('dmTTOption', c_short),
        ('dmCollate', c_short),
        ('dmFormName', BCHAR * CCHFORMNAME),
        ('dmLogPixels', WORD),
        ('dmBitsPerPel', DWORD),
        ('dmPelsWidth', DWORD),
        ('dmPelsHeight', DWORD),
        ('dmDisplayFlags', DWORD),  # union with dmNup
        ('dmDisplayFrequency', DWORD),
        ('dmICMMethod', DWORD),
        ('dmICMIntent', DWORD),
        ('dmDitherType', DWORD),
        ('dmReserved1', DWORD),
        ('dmReserved2', DWORD),
        ('dmPanningWidth', DWORD),
        ('dmPanningHeight', DWORD),
    ]


class ICONINFO(Structure):
    _fields_ = [
        ('fIcon', BOOL),
        ('xHotspot', DWORD),
        ('yHotspot', DWORD),
        ('hbmMask', HBITMAP),
        ('hbmColor', HBITMAP)
    ]
    __slots__ = [f[0] for f in _fields_]


class RAWINPUTDEVICE(Structure):
    _fields_ = [
        ('usUsagePage', USHORT),
        ('usUsage', USHORT),
        ('dwFlags', DWORD),
        ('hwndTarget', HWND)
    ]


PCRAWINPUTDEVICE = POINTER(RAWINPUTDEVICE)
HRAWINPUT = HANDLE


class RAWINPUTHEADER(Structure):
    _fields_ = [
        ('dwType', DWORD),
        ('dwSize', DWORD),
        ('hDevice', HANDLE),
        ('wParam', WPARAM),
    ]


class _Buttons(Structure):
    _fields_ = [
        ('usButtonFlags', USHORT),
        ('usButtonData', USHORT),
    ]


class _U(Union):
    _anonymous_ = ('_buttons',)
    _fields_ = [
        ('ulButtons', ULONG),
        ('_buttons', _Buttons),
    ]


class RAWMOUSE(Structure):
    _anonymous_ = ('u',)
    _fields_ = [
        ('usFlags', USHORT),
        ('u', _U),
        ('ulRawButtons', ULONG),
        ('lLastX', LONG),
        ('lLastY', LONG),
        ('ulExtraInformation', ULONG),
    ]


class RAWKEYBOARD(Structure):
    _fields_ = [
        ('MakeCode', USHORT),
        ('Flags', USHORT),
        ('Reserved', USHORT),
        ('VKey', USHORT),
        ('Message', UINT),
        ('ExtraInformation', ULONG),
    ]


class RAWHID(Structure):
    _fields_ = [
        ('dwSizeHid', DWORD),
        ('dwCount', DWORD),
        ('bRawData', POINTER(BYTE)),
    ]


class _RAWINPUTDEVICEUNION(Union):
    _fields_ = [
        ('mouse', RAWMOUSE),
        ('keyboard', RAWKEYBOARD),
        ('hid', RAWHID),
    ]


class RAWINPUT(Structure):
    _fields_ = [
        ('header', RAWINPUTHEADER),
        ('data', _RAWINPUTDEVICEUNION),
    ]


# PROPVARIANT wrapper, doesn't require InitPropVariantFromInt64 this way.
class _VarTable(Union):
    """Must be in an anonymous union or values will not work across various VT's."""
    _fields_ = [
        ('llVal', ctypes.c_longlong),
        ('pwszVal', LPWSTR)
    ]


class PROPVARIANT(Structure):
    _anonymous_ = ['union']

    _fields_ = [
        ('vt', ctypes.c_ushort),
        ('wReserved1', ctypes.c_ubyte),
        ('wReserved2', ctypes.c_ubyte),
        ('wReserved3', ctypes.c_ulong),
        ('union', _VarTable)
    ]


class _VarTableVariant(Union):
    """Must be in an anonymous union or values will not work across various VT's."""
    _fields_ = [
        ('bstrVal', LPCWSTR)
    ]


class VARIANT(Structure):
    _anonymous_ = ['union']

    _fields_ = [
        ('vt', ctypes.c_ushort),
        ('wReserved1', WORD),
        ('wReserved2', WORD),
        ('wReserved3', WORD),
        ('union', _VarTableVariant)
    ]


class DWM_BLURBEHIND(Structure):
    _fields_ = [
        ("dwFlags", DWORD),
        ("fEnable", BOOL),
        ("hRgnBlur", HRGN),
        ("fTransitionOnMaximized", DWORD),
    ]


class STATSTG(Structure):
    _fields_ = [
        ('pwcsName', LPOLESTR),
        ('type', DWORD),
        ('cbSize', ULARGE_INTEGER),
        ('mtime', FILETIME),
        ('ctime', FILETIME),
        ('atime', FILETIME),
        ('grfMode', DWORD),
        ('grfLocksSupported', DWORD),
        ('clsid', CLSID),
        ('grfStateBits', DWORD),
        ('reserved', DWORD),
    ]


class TIMECAPS(Structure):
    _fields_ = (('wPeriodMin', UINT),
                ('wPeriodMax', UINT))


class IStream(com.pIUnknown):
    _methods_ = [
        ('Read',
         com.STDMETHOD(c_void_p, ULONG, POINTER(ULONG))),
        ('Write',
         com.STDMETHOD()),
        ('Seek',
         com.STDMETHOD(LARGE_INTEGER, DWORD, POINTER(ULARGE_INTEGER))),
        ('SetSize',
         com.STDMETHOD()),
        ('CopyTo',
         com.STDMETHOD()),
        ('Commit',
         com.STDMETHOD()),
        ('Revert',
         com.STDMETHOD()),
        ('LockRegion',
         com.STDMETHOD()),
        ('UnlockRegion',
         com.STDMETHOD()),
        ('Stat',
         com.STDMETHOD(POINTER(STATSTG), UINT)),
        ('Clone',
         com.STDMETHOD()),
    ]


class DEV_BROADCAST_HDR(Structure):
    _fields_ = (
        ('dbch_size', DWORD),
        ('dbch_devicetype', DWORD),
        ('dbch_reserved', DWORD),
    )


class DEV_BROADCAST_DEVICEINTERFACE(Structure):
    _fields_ = (
        ('dbcc_size', DWORD),
        ('dbcc_devicetype', DWORD),
        ('dbcc_reserved', DWORD),
        ('dbcc_classguid', com.GUID),
        ('dbcc_name', ctypes.c_wchar * 256)
    )


class DISPLAY_DEVICEW(ctypes.Structure):
    _fields_ = [
        ('cb', DWORD),
        ('DeviceName', WCHAR * 32),
        ('DeviceString', WCHAR * 128),
        ('StateFlags', DWORD),
        ('DeviceID', WCHAR * 128),
        ('DeviceKey', WCHAR * 128),
    ]


# Structures below are to retrieve a monitor name...
class LUID(ctypes.Structure):
    _fields_ = [('LowPart', DWORD), ('HighPart', LONG)]


class _SourceInfoStruct(ctypes.Structure):
    _fields_ = [('cloneGroupId', UINT32, 16), ('sourceModeInfoIdx', UINT32, 16)]


class _DisplayUnion(ctypes.Union):
    _fields_ = [('modeInfoIdx', UINT32), ('DUMMYSTRUCTNAME', _SourceInfoStruct)]


class DISPLAYCONFIG_PATH_SOURCE_INFO(ctypes.Structure):
    _fields_ = [('adapterId', LUID), ('id', UINT32), ('DUMMYUNIONNAME', _DisplayUnion), ('statusFlags', UINT32)]


class _DummyStructTarget(ctypes.Structure):
    _fields_ = [
        ('desktopModeInfoIdx', UINT32, 16),
        ('targetModeInfoIdx', UINT32, 16),
    ]


class _DummyUnionTarget(ctypes.Union):
    _fields_ = [('modeInfoIdx', UINT32), ('DUMMYSTRUCTNAME', _DummyStructTarget)]


class DISPLAYCONFIG_RATIONAL(ctypes.Structure):
    _fields_ = [('Numerator', UINT32), ('Denominator', UINT32)]

    def __repr__(self):
        return f"DISPLAYCONFIG_RATIONAL(num={self.Numerator}, denom={self.Denominator})"


class DISPLAYCONFIG_PATH_TARGET_INFO(ctypes.Structure):
    _fields_ = [
        ('adapterId', LUID),
        ('id', UINT32),
        ('DUMMYUNIONNAME', _DummyUnionTarget),
        ('outputTechnology', UINT32),
        ('rotation', UINT32),
        ('scaling', UINT32),
        ('refreshRate', DISPLAYCONFIG_RATIONAL),
        ('scanLineOrdering', UINT32),
        ('targetAvailable', BOOL),
        ('statusFlags', UINT32),
    ]


class DISPLAYCONFIG_PATH_INFO(ctypes.Structure):
    _fields_ = [
        ('sourceInfo', DISPLAYCONFIG_PATH_SOURCE_INFO),
        ('targetInfo', DISPLAYCONFIG_PATH_TARGET_INFO),
        ('flags', UINT32),
    ]


class DISPLAYCONFIG_DEVICE_INFO_HEADER(ctypes.Structure):
    _fields_ = [('type', UINT32),
                ('size', UINT32),
                ('adapterId', LUID),
                ('id', UINT32)
                ]


class DISPLAYCONFIG_SOURCE_DEVICE_NAME(ctypes.Structure):
    _fields_ = [
        ('header', DISPLAYCONFIG_DEVICE_INFO_HEADER),
        ('viewGdiDeviceName', WCHAR * 32)
    ]


class DISPLAYCONFIG_TARGET_DEVICE_NAME(ctypes.Structure):
    _fields_ = [
        ('header', DISPLAYCONFIG_DEVICE_INFO_HEADER),
        ('flags', UINT32),
        ('outputTechnology', UINT32),
        ('edidManufactureId', UINT16),
        ('edidProductCodeId', UINT16),
        ('connectorInstance', UINT32),
        ('monitorFriendlyDeviceName', WCHAR * 64),
        ('monitorDevicePath', WCHAR * 128),
    ]
